# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional


def packages(exclude_packages: Optional[list[str]] = None) -> list[str]:
    """Get list of project packages"""
    _exclude_packages = set(exclude_packages or [])
    all_packages = _flatten_packages(
        _get_package_dict(f"{os.path.dirname(__file__)}/mlrun"), parent_key="mlrun"
    )
    return list(sorted(all_packages.difference(_exclude_packages)))


def _get_package_dict(starting_path, exclude: Optional[list[str]] = None) -> dict:
    """Get hierarchical dict of packages from starting path"""
    package_dict = {}
    exclude = exclude or ["__pycache__"]

    for dir_path, dir_names, _ in os.walk(starting_path):
        key_path = dir_path.replace(starting_path, "")
        sub_package_dict = package_dict
        for sub_package in key_path.split("/"):
            if sub_package and sub_package not in exclude:
                sub_package_dict = sub_package_dict[sub_package]

        for dir_name in dir_names:
            if dir_name not in exclude:
                sub_package_dict[dir_name] = {}

    return package_dict


def _flatten_packages(
    package_dict: dict, parent_key: str = "", sep: str = "."
) -> set[str]:
    """Flatten hierarchical dict of packages to set of packages"""
    items = [parent_key] if parent_key else []
    for key, value in package_dict.items():
        new_key = f"{parent_key}{sep}{key}" if parent_key else key
        items.append(new_key)
        if isinstance(value, dict) and value:
            items.extend(_flatten_packages(value, new_key, sep=sep))

    return set(items)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional, Union

import torch
from torch import Tensor

import mlrun
from mlrun.artifacts import Artifact

from ..._common import LoggingMode
from ..._dl_common.loggers import MLRunLogger
from ..model_handler import PyTorchModelHandler
from ..utils import PyTorchTypes
from .logging_callback import LoggingCallback


class MLRunLoggingCallback(LoggingCallback):
    """
    Callback for logging data during training / validation via mlrun's context. Each tracked hyperparameter and metrics
    results will be logged per epoch and at the end of the run the model will be saved and logged as well. Some plots
    will be available as well. To summarize, the available data in mlrun will be:

    * For each epoch:

      * Tracking table: epoch, static hyperparameters, dynamic hyperparameters, training metrics, validation metrics.
      * Per iteration (batch) chart artifacts for the training and validation metrics.

    * At the end of the run:

      * Per epoch chart artifacts for the validation summaries and dynamic hyperparameters.
      * Model is logged with all of the files and artifacts.

    All the collected data will be available in this callback post the training / validation process and can be accessed
    via the 'training_results', 'validation_results', 'static_hyperparameters', 'dynamic_hyperparameters' and
    'summaries' properties.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx,
        model_handler: PyTorchModelHandler,
        log_model_tag: str = "",
        log_model_labels: Optional[dict[str, PyTorchTypes.TrackableType]] = None,
        log_model_parameters: Optional[dict[str, PyTorchTypes.TrackableType]] = None,
        log_model_extra_data: Optional[
            dict[str, Union[PyTorchTypes.TrackableType, Artifact]]
        ] = None,
        dynamic_hyperparameters: Optional[
            dict[
                str,
                tuple[
                    str,
                    Union[
                        list[Union[str, int]], Callable[[], PyTorchTypes.TrackableType]
                    ],
                ],
            ]
        ] = None,
        static_hyperparameters: Optional[
            dict[
                str,
                Union[PyTorchTypes.TrackableType, tuple[str, list[Union[str, int]]]],
            ]
        ] = None,
        auto_log: bool = False,
    ):
        """
        Initialize an mlrun logging callback with the given hyperparameters and logging configurations. Notice: In order
        to log the model, its class (torch.Module) must be in the custom objects map or the modules map.

        :param context:                  MLRun context to log to. Its parameters will be logged automatically  if
                                         'auto_log' is True.
        :param model_handler:            The model handler to use for logging the model at the end of the run with the
                                         collected logs.
        :param log_model_tag:            Version tag to give the logged model.
        :param log_model_labels:         Labels to log with the model.
        :param log_model_parameters:     Parameters to log with the model.
        :param log_model_extra_data:     Extra data to log with the model.
        :param dynamic_hyperparameters:  If needed to track a hyperparameter dynamically (sample it each epoch) it
                                         should be passed here. The parameter expects a dictionary where the keys are
                                         the hyperparameter chosen names and the values are tuples of object key and a
                                         list with the key chain. A key chain is a list of keys and indices to know how
                                         to access the needed hyperparameter. If the hyperparameter is not of accessible
                                         from any of the HyperparametersKeys, a custom callable method can be passed in
                                         the tuple instead of the key chain when providing the word
                                         HyperparametersKeys.CUSTOM. For example, to track the 'lr' attribute of
                                         an optimizer and a custom parameter, one should pass:
                                         {
                                             "lr": (HyperparametersKeys.OPTIMIZER, ["param_groups", 0, "lr"]),
                                             "custom parameter": (HyperparametersKeys.CUSTOM, get_custom_parameter)
                                         }
        :param static_hyperparameters:   If needed to track a hyperparameter one time per run it should be passed here.
                                         The parameter expects a dictionary where the keys are the
                                         hyperparameter chosen names and the values are the hyperparameter static value
                                         or a tuple of object key and a list with the key chain just like the dynamic
                                         hyperparameter. For example, to track the 'epochs' of an experiment run, one
                                         should pass:
                                         {
                                             "epochs": 7
                                         }
        :param auto_log:                 Whether or not to enable auto logging for logging the context parameters and
                                         trying to track common static and dynamic hyperparameters.
        """
        super().__init__(
            dynamic_hyperparameters=dynamic_hyperparameters,
            static_hyperparameters=static_hyperparameters,
            auto_log=auto_log,
        )

        # Replace the logger with an MLRunLogger:
        del self._logger
        self._logger = MLRunLogger(context=context)

        # Store the given handler:
        self._model_handler = model_handler

        # Store the attributes to log along the model:
        self._log_model_tag = log_model_tag
        self._log_model_labels = log_model_labels
        self._log_model_parameters = log_model_parameters
        self._log_model_extra_data = log_model_extra_data

        # Setup the additional PyTorchModelHandler parameters for logging the model later:
        self._input_sample = None  # type: PyTorchModelHandler.IOSample
        self._output_sample = None  # type: PyTorchModelHandler.IOSample

    def on_run_end(self):
        """
        Before the run ends, this method will be called to log the model and the run summaries charts.
        """
        # Check if the logger is in evaluation mode, if so, log the last epoch
        if self._logger.mode == LoggingMode.EVALUATION:
            self._logger.log_epoch_to_context(epoch=1)

        # Set the inputs and outputs:
        if self._model_handler.inputs is None:
            self._model_handler.set_inputs(from_sample=self._input_sample)
        if self._model_handler.outputs is None:
            self._model_handler.set_outputs(from_sample=self._output_sample)

        # End the run:
        self._logger.log_run(
            model_handler=self._model_handler,
            tag=self._log_model_tag,
            labels=self._log_model_labels,
            parameters=self._log_model_parameters,
            extra_data=self._log_model_extra_data,
        )

    def on_epoch_end(self, epoch: int):
        """
        Before the given epoch ends, this method will be called to log the dynamic hyperparameters and results of this
        epoch via the stored context.

        :param epoch: The epoch that has just ended.
        """
        super().on_epoch_end(epoch=epoch)

        # Create child context to hold the current epoch's results:
        self._logger.log_epoch_to_context(epoch=epoch)

    def on_inference_begin(self, x: Tensor):
        """
        Before the inference of the current batch sample into the model, this method will be called to save an input
        sample - a zeros tensor with the same properties of the 'x' input.

        :param x: The input of the current batch.
        """
        if self._input_sample is None:
            self._input_sample = torch.zeros(size=x.shape, dtype=x.dtype)

    def on_inference_end(self, y_pred: Tensor, y_true: Tensor):
        """
        After the inference of the current batch sample, this method will be called to save an output sample - a zeros
        tensor with the same properties of the 'y_pred' output.

        :param y_pred: The prediction (output) of the model for this batch's input ('x').
        :param y_true: The ground truth value of the current batch.
        """
        if self._output_sample is None:
            self._output_sample = torch.zeros(size=y_pred.shape, dtype=y_pred.dtype)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from abc import ABC
from enum import Enum
from pathlib import Path
from typing import Any, TypeVar, Union

import numpy as np
import pandas as pd

import mlrun
from mlrun.artifacts import Artifact
from mlrun.data_types import ValueType
from mlrun.datastore import DataItem


class CommonTypes(ABC):
    """
    Common type hints to all frameworks.
    """

    # A generic model type in a handler / interface (examples: tf.keras.Model, torch.Module):
    ModelType = TypeVar("ModelType")

    # A generic input / output samples for reading the inputs / outputs properties:
    IOSampleType = TypeVar("IOSampleType")

    # A generic object type for what can be wrapped with a framework MLRun interface (examples: xgb, xgb.XGBModel):
    MLRunInterfaceableType = TypeVar("MLRunInterfaceableType")

    # Type for a MLRun Interface restoration tuple as returned from 'remove_interface':
    MLRunInterfaceRestorationType = tuple[
        dict[str, Any],  # Interface properties.
        dict[str, Any],  # Replaced properties.
        list[str],  # Replaced methods and functions.
    ]

    # Common dataset type to all frameworks:
    DatasetType = Union[
        list,
        tuple,
        dict,
        np.ndarray,
        pd.DataFrame,
        pd.Series,
        "scipy.sparse.base.spmatrix",  # noqa: F821
    ]

    # A joined type for receiving a path from 'pathlib' or 'os.path':
    PathType = Union[str, Path]

    # A joined type for all trackable values (for logging):
    TrackableType = Union[str, bool, float, int]

    # Types available in the extra data dictionary of an artifact:
    ExtraDataType = Union[str, bytes, Artifact, DataItem]


class LoggingMode(Enum):
    """
    The logging mode options.
    """

    TRAINING = "Training"
    EVALUATION = "Evaluation"


class CommonUtils(ABC):
    """
    Common utilities functions to all frameworks.
    """

    @staticmethod
    def to_array(dataset: CommonTypes.DatasetType) -> np.ndarray:
        """
        Convert the given dataset to np.ndarray.

        :param dataset: The dataset to convert. Must be one of {pd.DataFrame, pd.Series, scipy.sparse.base.spmatrix,
                        list, tuple, dict}.

        :return: The dataset as a ndarray.

        :raise MLRunInvalidArgumentError: If the dataset type is not supported.
        """
        if isinstance(dataset, np.ndarray):
            return dataset
        if isinstance(dataset, (pd.DataFrame, pd.Series)):
            return dataset.to_numpy()
        if isinstance(dataset, (list, tuple)):
            return np.array(dataset)
        if isinstance(dataset, dict):
            return np.array(list(dataset.values()))
        try:
            # SciPy is not in MLRun's requirements but common to all frameworks.
            import scipy.sparse.base as sp

            if isinstance(dataset, sp.spmatrix):
                return dataset.toarray()
        except ModuleNotFoundError:
            # SciPy is not installed.
            pass

        raise mlrun.errors.MLRunInvalidArgumentError(
            f"Could not convert the given dataset into a numpy ndarray. Supporting conversion from: "
            f"{CommonUtils.get_union_typehint_string(CommonTypes.DatasetType)}. "
            f"The given dataset was of type: '{type(dataset)}'"
        )

    @staticmethod
    def to_dataframe(dataset: CommonTypes.DatasetType) -> pd.DataFrame:
        """
        Convert the given dataset to pd.DataFrame.

        :param dataset: The dataset to convert. Must be one of {np.ndarray, pd.Series, scipy.sparse.base.spmatrix, list,
                        tuple, dict}.

        :return: The dataset as a DataFrame.

        :raise MLRunInvalidArgumentError: If the dataset type is not supported.
        """
        if isinstance(dataset, pd.DataFrame):
            return dataset
        if isinstance(dataset, (np.ndarray, pd.Series, list, tuple, dict)):
            return pd.DataFrame(dataset)
        try:
            # SciPy is not in MLRun's requirements but common to all frameworks.
            import scipy.sparse.base as sp

            if isinstance(dataset, sp.spmatrix):
                return pd.DataFrame.sparse.from_spmatrix(dataset)
        except ModuleNotFoundError:
            # SciPy is not installed.
            pass
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"Could not convert the given dataset into a pandas DataFrame. Supporting conversion from: "
            f"{CommonUtils.get_union_typehint_string(CommonTypes.DatasetType)}. "
            f"The given dataset was of type: '{type(dataset)}'"
        )

    @staticmethod
    def convert_value_type_to_np_dtype(
        value_type: str,
    ) -> np.dtype:
        """
        Get the 'numpy.dtype' equivalent to the given MLRun value type.

        :param value_type: The MLRun value type to convert to numpy data type.

        :return: The 'numpy.dtype' equivalent to the given MLRun data type.

        :raise MLRunInvalidArgumentError: If numpy is not supporting the given data type.
        """
        # Initialize the mlrun to numpy data type conversion map:
        conversion_map = {
            ValueType.BOOL: bool,
            ValueType.INT8: np.int8,
            ValueType.INT16: np.int16,
            ValueType.INT32: np.int32,
            ValueType.INT64: np.int64,
            ValueType.UINT8: np.uint8,
            ValueType.UINT16: np.uint16,
            ValueType.UINT32: np.uint32,
            ValueType.UINT64: np.uint64,
            ValueType.FLOAT16: np.float16,
            ValueType.FLOAT: np.float32,
            ValueType.DOUBLE: np.float64,
        }

        # Convert and return:
        if value_type in conversion_map:
            return conversion_map[value_type]
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"The ValueType given is not supported in numpy: '{value_type}'"
        )

    @staticmethod
    def convert_np_dtype_to_value_type(np_dtype: Union[np.dtype, type, str]) -> str:
        """
        Convert the given numpy data type to MLRun value type. It is better to use explicit bit namings (for example:
        instead of using 'np.double', use 'np.float64').

        :param np_dtype: The numpy data type to convert to MLRun's value type. Expected to be a 'numpy.dtype', 'type' or
                         'str'.

        :return: The MLRun value type converted from the given data type.

        :raise MLRunInvalidArgumentError: If the numpy data type is not supported by MLRun.
        """
        # Initialize the numpy to mlrun data type conversion map:
        conversion_map = {
            bool.__name__: ValueType.BOOL,
            np.byte.__name__: ValueType.INT8,
            np.int8.__name__: ValueType.INT8,
            np.short.__name__: ValueType.INT16,
            np.int16.__name__: ValueType.INT16,
            np.int32.__name__: ValueType.INT32,
            int.__name__: ValueType.INT64,
            np.int64.__name__: ValueType.INT64,
            np.ubyte.__name__: ValueType.UINT8,
            np.uint8.__name__: ValueType.UINT8,
            np.ushort.__name__: ValueType.UINT16,
            np.uint16.__name__: ValueType.UINT16,
            np.uint32.__name__: ValueType.UINT32,
            np.uint.__name__: ValueType.UINT64,
            np.uint64.__name__: ValueType.UINT64,
            np.half.__name__: ValueType.FLOAT16,
            np.float16.__name__: ValueType.FLOAT16,
            np.single.__name__: ValueType.FLOAT,
            np.float32.__name__: ValueType.FLOAT,
            np.double.__name__: ValueType.DOUBLE,
            float.__name__: ValueType.DOUBLE,
            np.float64.__name__: ValueType.DOUBLE,
        }

        # Parse the given numpy data type to string:
        if isinstance(np_dtype, np.dtype):
            np_dtype = np_dtype.name
        elif isinstance(np_dtype, type):
            np_dtype = np_dtype.__name__

        # Convert and return:
        if np_dtype in conversion_map:
            return conversion_map[np_dtype]
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"MLRun value type is not supporting the given numpy data type: '{np_dtype}'"
        )

    @staticmethod
    def get_union_typehint_string(union_typehint) -> str:
        """
        Get the string representation of a types.Union typehint object.

        :param union_typehint: The union typehint to get its string representation.

        :return: The union typehint's string.
        """
        return re.sub(r"typing.Union|[\[\]'\"()]|ForwardRef", "", str(union_typehint))



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mlrun
from mlrun.datastore.store_resources import ResourceCache
from mlrun.datastore.targets import get_online_target
from mlrun.serving.server import create_graph_server

from ..feature_vector_utils import OnlineVectorService
from .base import BaseMerger


class StoreyFeatureMerger(BaseMerger):
    engine = "storey"
    support_online = True

    def __init__(self, vector, **engine_args):
        super().__init__(vector, **engine_args)
        self.impute_policy = engine_args.get("impute_policy")

    def _generate_online_feature_vector_graph(
        self,
        entity_keys,
        feature_set_fields,
        feature_set_objects,
        fixed_window_type,
    ):
        graph = self.vector.spec.graph.copy()
        start_states, default_final_state, responders = graph.check_and_process_graph(
            allow_empty=True
        )
        next = graph

        join_graph = self._get_graph(
            feature_set_objects, feature_set_fields, entity_keys
        )

        all_columns = []
        save_column = []
        entity_keys = []
        del_columns = []
        end_aliases = {}
        for step in join_graph.steps:
            name = step.right_feature_set_name
            feature_set = feature_set_objects[name]
            columns = feature_set_fields[name]
            column_names = [name for name, alias in columns]
            aliases = {name: alias for name, alias in columns if alias}
            all_columns += [aliases.get(name, name) for name in column_names]
            saved_columns_for_relation = list(
                self.vector.get_feature_set_relations(feature_set).keys()
            )

            for col in saved_columns_for_relation:
                if col not in column_names:
                    column_names.append(col)
                    del_columns.append(col)
                else:
                    save_column.append(col)

            entity_list = step.right_keys or list(feature_set.spec.entities.keys())
            if not entity_keys:
                # if entity_keys not provided by the user we will set it to be the entity of the first feature set
                entity_keys = entity_list
            end_aliases.update(
                {
                    k: v
                    for k, v in zip(entity_list, step.left_keys)
                    if k != v and v in save_column
                }
            )
            mapping = {k: v for k, v in zip(step.left_keys, entity_list) if k != v}
            if mapping:
                next = next.to(
                    "storey.Rename",
                    f"rename-{name}",
                    mapping=mapping,
                )

            next = next.to(
                "storey.QueryByKey",
                f"query-{name}",
                features=column_names,
                table=feature_set.uri,
                key_field=entity_list,
                aliases=aliases,
                fixed_window_type=fixed_window_type.to_qbk_fixed_window_type(),
            )
        if end_aliases:
            # run if the user want to save a column that related to another entity
            next = next.to(
                "storey.Rename",
                "rename-entity-to-features",
                mapping=end_aliases,
            )
        if del_columns:
            next = next.to(
                "storey.flow.DropColumns",
                "drop-unnecessary-columns",
                columns=del_columns,
            )
        for name in start_states:
            next.set_next(name)

        if not start_states:  # graph was empty
            next.respond()
        elif not responders and default_final_state:  # graph has clear state sequence
            graph[default_final_state].respond()
        elif not responders:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "the graph doesnt have an explicit final step to respond on"
            )
        return graph, all_columns, entity_keys

    def init_online_vector_service(
        self, entity_keys, fixed_window_type, update_stats=False
    ):
        try:
            from storey import SyncEmitSource
        except ImportError as exc:
            raise ImportError(f"storey not installed, use pip install storey, {exc}")

        feature_set_objects, feature_set_fields = self.vector.parse_features(
            offline=False, update_stats=update_stats
        )
        if not feature_set_fields:
            raise mlrun.errors.MLRunRuntimeError(
                f"No features found for feature vector '{self.vector.metadata.name}'"
            )
        (
            graph,
            requested_columns,
            entity_keys,
        ) = self._generate_online_feature_vector_graph(
            entity_keys,
            feature_set_fields,
            feature_set_objects,
            fixed_window_type,
        )
        graph.set_flow_source(SyncEmitSource())
        server = create_graph_server(graph=graph, parameters={})

        cache = ResourceCache()
        for featureset in feature_set_objects.values():
            driver = get_online_target(featureset)
            if not driver:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"resource {featureset.uri} does not have an online data target"
                )
            cache.cache_table(featureset.uri, driver.get_table_object())

        server.init_states(context=None, namespace=None, resource_cache=cache)
        server.init_object(None)

        service = OnlineVectorService(
            self.vector,
            graph,
            entity_keys,
            impute_policy=self.impute_policy,
            requested_columns=requested_columns,
        )
        service.initialize()

        return service



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Optional


def base_requirements() -> list[str]:
    return list(_load_dependencies_from_file("requirements.txt"))


def extra_requirements() -> dict[str, list[str]]:
    # NOTE:
    #     - These are tested in `automation/package_test/test.py`. If you modify these, make sure to change the
    #       corresponding line there.
    #     - We have a copy of these in extras-requirements.txt. If you modify these, make sure to change it
    #       there as well
    extras_require = {
        "s3": [
            "boto3>=1.28.0,<1.36",
            "aiobotocore>=2.5.0,<2.16",
            "s3fs>=2025.5.1, <=2025.7.0",
        ],
        "azure-blob-storage": [
            "msrest~=0.6.21",
            "azure-core~=1.24",
            "adlfs==2024.12.0",
            "pyopenssl>=23",
        ],
        "azure-key-vault": [
            "azure-identity~=1.5",
            "azure-keyvault-secrets~=4.2",
            "pyopenssl>=23",
        ],
        "plotly": ["plotly~=5.23"],
        # used to generate visualization nuclio/serving graph steps
        "graphviz": ["graphviz~=0.20.0"],
        "google-cloud": [
            "google-cloud-storage==2.14.0",
            "google-cloud-bigquery[pandas, bqstorage]==3.14.1",
            # google-cloud-bigquery[bqstorage] requires google-cloud-bigquery-storage >= 2.6.0, but older (<2.17)
            # versions of google-cloud-bigquery-storage runs into an issue
            # (https://github.com/pypa/setuptools/issues/4476) with setuptools (ML-7273)
            "google-cloud-bigquery-storage~=2.17",
            "google-cloud==0.34",
            "gcsfs>=2025.5.1, <=2025.7.0",
        ],
        "kafka": [
            "kafka-python~=2.1.0",
            # because confluent kafka supports avro format by default
            "avro~=1.11",
        ],
        "redis": ["redis~=4.3"],
        "mlflow": ["mlflow~=2.22"],
        "databricks-sdk": ["databricks-sdk~=0.20.0"],
        "sqlalchemy": ["sqlalchemy~=2.0"],
        "dask": [
            # Use ~= instead of >= to avoid installing newer versions of dask and distributed,
            # which can cause incompatibilities between the client and the Dask scheduler/worker.
            # (both must be the same version)
            # Reference: https://blog.dask.org/2023/04/14/scheduler-environment-requirements
            # Note: dask 2023 does not work on Python 3.11, and dask 2024 requires dependencies
            # that MLRun with Python 3.9 cannot support.
            'dask~=2023.12.1; python_version < "3.11"',
            'dask==2024.8; python_version >= "3.11"',
            'distributed~=2023.12.1; python_version < "3.11"',
            'distributed==2024.8; python_version >= "3.11"',
        ],
        "alibaba-oss": ["ossfs==2025.5.0", "oss2==2.18.4"],
        "tdengine": ["taos-ws-py==0.3.2"],
        "snowflake": ["snowflake-connector-python~=3.7"],
    }

    api_deps = list(
        _load_dependencies_from_file("dockerfiles/mlrun-api/requirements.txt")
    )
    extras_require.update(
        {
            "dev-postgres": ["pytest-mock-resources[postgres]~=2.12"],
            "kfp18": ["mlrun_pipelines_kfp_v1_8[kfp]~=0.5.8"],
            # TODO uncomment when KFP 1.8 support is removed
            # "kfp2": ["mlrun_pipelines_kfp_v2[kfp]>=0.5.0 ; python_version >= '3.11'"],
            "api": api_deps,
            "all": _get_extra_dependencies(extras_require=extras_require),
            "complete": _get_extra_dependencies(
                extras_require=extras_require,
            ),
            "complete-api": _get_extra_dependencies(
                base_deps=api_deps,
                extras_require=extras_require,
            ),
        }
    )
    return extras_require


def _is_ignored(line: str) -> bool:
    line = line.strip()
    return (not line) or (line[0] == "#") or line.startswith("git+")


def _extract_package_from_egg(line: str) -> str:
    """Extract egg name from line"""
    if "#egg=" in line:
        _, package = line.split("#egg=")
        return f"{package} @ {line}"
    return line


def _load_dependencies_from_file(
    path: str, parent_dir: Optional[str] = None
) -> list[str]:
    """Load dependencies from requirements file"""
    parent_dir = parent_dir or os.path.dirname(__file__)
    with open(f"{parent_dir}/{path}") as fp:
        return [
            _extract_package_from_egg(line.strip())
            for line in fp
            if not _is_ignored(line)
        ]


def _get_extra_dependencies(
    include: Optional[list[str]] = None,
    exclude: Optional[list[str]] = None,
    base_deps: Optional[list[str]] = None,
    extras_require: Optional[dict[str, list[str]]] = None,
) -> list[str]:
    """Get list of dependencies for given extras categories

    :param include: list of extras categories to include
    :param exclude: list of extras categories to exclude
    :param base_deps: list of base dependencies to include
    :return: list of dependencies
    """
    include = include or []
    exclude = exclude or []
    base_deps = base_deps or []
    extras_require = extras_require or {}
    extra_deps = set(base_deps)
    for extra_key, requirement_list in extras_require.items():
        if extra_key not in exclude and (not include or extra_key in include):
            extra_deps.update(requirement_list)
    return list(sorted(extra_deps))



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from typing import Optional

import mlrun
import mlrun.common.constants as mlrun_constants
from mlrun.config import config as mlconf
from mlrun.model import DataTargetBase, new_task
from mlrun.runtimes.function_reference import FunctionReference
from mlrun.utils import logger

from ...runtimes import RuntimeKinds
from ..common import RunConfig
from .base import BaseMerger


def run_merge_job(
    vector,
    target: DataTargetBase,
    merger: BaseMerger,
    engine: str,
    engine_args: dict,
    spark_service: Optional[str] = None,
    entity_rows=None,
    entity_timestamp_column=None,
    run_config=None,
    drop_columns=None,
    with_indexes=None,
    query=None,
    order_by=None,
    start_time=None,
    end_time=None,
    timestamp_for_filtering=None,
    additional_filters=None,
):
    name = vector.metadata.name
    if not target or not hasattr(target, "to_dict"):
        raise mlrun.errors.MLRunInvalidArgumentError("target object must be specified")
    name = f"{name}-merger"
    run_config = run_config or RunConfig()
    kind = run_config.kind or ("spark" if engine == "spark" else "job")
    run_config.kind = kind
    default_code = _default_merger_handler.replace("{{{engine}}}", merger.__name__)
    if not run_config.function:
        function_ref = vector.spec.function.copy()
        if function_ref.is_empty():
            function_ref = FunctionReference(name=name, kind=kind)
        if not function_ref.url:
            function_ref.code = default_code
        run_config.function = function_ref

    function = run_config.to_function(kind, merger.get_default_image(kind))

    # Avoid overriding a handler that was provided by the user
    if not run_config.handler:
        function.with_code(body=default_code)
    else:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "get_offline_features does not support run_config with a handler"
        )

    function.metadata.project = vector.metadata.project
    function.metadata.name = function.metadata.name or name

    if run_config.kind == RuntimeKinds.remotespark:
        if not spark_service:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "spark_service must be set when running with a remote spark runtime"
            )
        function.with_spark_service(spark_service=spark_service)
    elif run_config.kind == RuntimeKinds.spark:
        if mlconf.is_running_on_iguazio():
            function.with_igz_spark()

        def set_default_resources(resources, setter_function):
            requests = resources.get("requests")
            set_memory = requests is None or "memory" not in requests
            set_cpu = requests is None or "cpu" not in requests
            if set_memory or set_cpu:
                mem = "1G" if set_memory else None
                cpu = "1" if set_cpu else None
                setter_function(mem=mem, cpu=cpu, patch=True)

        set_default_resources(
            function.spec.driver_resources, function.with_driver_requests
        )
        set_default_resources(
            function.spec.executor_resources, function.with_executor_requests
        )
    if start_time and not isinstance(start_time, str):
        start_time = start_time.isoformat()
    if end_time and not isinstance(end_time, str):
        end_time = end_time.isoformat()

    task = new_task(
        name=name,
        params={
            "vector_uri": vector.uri,
            "target": target.to_dict(),
            "entity_timestamp_column": entity_timestamp_column,
            "drop_columns": drop_columns,
            "with_indexes": with_indexes,
            "query": query,
            "order_by": order_by,
            "start_time": start_time,
            "end_time": end_time,
            "timestamp_for_filtering": timestamp_for_filtering,
            "engine_args": engine_args,
            "additional_filters": additional_filters,
        },
        inputs={"entity_rows": entity_rows} if entity_rows is not None else {},
    )
    task.spec.secret_sources = run_config.secret_sources
    task.set_label(
        mlrun_constants.MLRunInternalLabels.job_type, "feature-merge"
    ).set_label(mlrun_constants.MLRunInternalLabels.feature_vector, vector.uri)
    task.metadata.uid = uuid.uuid4().hex
    vector.status.run_uri = task.metadata.uid
    vector.save()

    run = function.run(
        task,
        handler=run_config.handler or "merge_handler",
        local=run_config.local,
        watch=run_config.watch,
    )
    logger.info(f"Feature vector merge job started, run id = {run.uid()}")
    return RemoteVectorResponse(vector, run, with_indexes, drop_columns)


class RemoteVectorResponse:
    """get_offline_features response object"""

    def __init__(self, vector, run, with_indexes=False, drop_columns=None):
        self.run = run
        self.vector = vector
        self.with_indexes = with_indexes or self.vector.spec.with_indexes
        self.drop_columns = drop_columns

    @property
    def status(self):
        """vector prep job status (ready, running, error)"""
        return self.run.state()

    def _is_ready(self):
        if self.status != "completed":
            raise mlrun.errors.MLRunTaskNotReadyError(
                "feature vector dataset is not ready"
            )
        self.vector.reload()

    def to_dataframe(self, columns=None, df_module=None, **kwargs):
        """return result as a dataframe object (generated from the dataitem).

        :param columns:   optional, list of columns to select
        :param df_module: optional, py module used to create the DataFrame (e.g. pd, dd, cudf, ..)
        :param kwargs:    extended DataItem.as_df() args
        """
        self._is_ready()
        if not columns:
            columns = list(self.vector.status.features.keys())
            if self.with_indexes:
                columns += self.vector.status.index_keys
                if self.vector.status.timestamp_key is not None:
                    columns.insert(0, self.vector.status.timestamp_key)
            if self.drop_columns:
                for drop_col in self.drop_columns:
                    if drop_col in columns:
                        columns.remove(drop_col)

        file_format = kwargs.get("format")
        if not file_format:
            file_format = self.run.status.results["target"]["kind"]

        df = mlrun.get_dataitem(self.target_uri).as_df(
            columns=columns, df_module=df_module, format=file_format, **kwargs
        )
        if self.with_indexes:
            df.set_index(self.vector.status.index_keys, inplace=True, drop=True)
        return df

    @property
    def target_uri(self):
        """return path of the results file"""
        self._is_ready()
        return self.run.output("target")["path"]


_default_merger_handler = """
import mlrun
import mlrun.feature_store.retrieval
from mlrun.datastore.targets import get_target_driver
def merge_handler(context, vector_uri, target, entity_rows=None, 
                  entity_timestamp_column=None, drop_columns=None, with_indexes=None, query=None,
                  engine_args=None, order_by=None, start_time=None, end_time=None, timestamp_for_filtering=None,
                  additional_filters=None):
    vector = context.get_store_resource(vector_uri)
    store_target = get_target_driver(target, vector)
    if entity_rows:
        entity_rows = entity_rows.as_df()

    context.logger.info(f"Starting vector merge task to {vector.uri}")
    merger = mlrun.feature_store.retrieval.{{{engine}}}(vector, **(engine_args or {}))
    merger.start(entity_rows, entity_timestamp_column, store_target, drop_columns, with_indexes=with_indexes, 
                 query=query, order_by=order_by, start_time=start_time, end_time=end_time,
                 timestamp_for_filtering=timestamp_for_filtering, additional_filters=additional_filters)

    target = vector.status.targets[store_target.name].to_dict()
    context.log_result('feature_vector', vector.uri)
    context.log_result('target', target)
"""  # noqa



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing

import kubernetes.client

import mlrun.common.schemas.function
import mlrun.errors
import mlrun.k8s_utils
import mlrun.runtimes.pod
from mlrun.config import config
from mlrun.runtimes.mounts import mount_v3io, mount_v3iod

from ...execution import MLClientCtx
from ...model import RunObject
from ...utils import update_in, verify_field_regex
from ..kubejob import KubejobRuntime
from ..pod import KubeResourceSpec
from ..utils import (
    generate_resources,
    get_gpu_from_resource_requirement,
    get_item_name,
    verify_limits,
    verify_requests,
)


class Spark3JobSpec(KubeResourceSpec):
    _jvm_memory_resource_notation = r"^[0-9]+[KkMmGg]$"

    # https://github.com/GoogleCloudPlatform/spark-on-k8s-operator/blob/55732a6a392cbe1d6546c7ec6823193ab055d2fa/pkg/apis/sparkoperator.k8s.io/v1beta2/types.go#L181
    _dict_fields = KubeResourceSpec._dict_fields + [
        "driver_resources",
        "executor_resources",
        "job_type",
        "python_version",
        "spark_version",
        "restart_policy",
        "deps",
        "main_class",
        "spark_conf",
        "hadoop_conf",
        "use_default_image",
        "monitoring",
        "driver_node_selector",
        "executor_node_selector",
        "dynamic_allocation",
        "driver_tolerations",
        "executor_tolerations",
        "driver_affinity",
        "executor_affinity",
        "driver_preemption_mode",
        "executor_preemption_mode",
        "driver_volume_mounts",
        "executor_volume_mounts",
        "driver_java_options",
        "executor_java_options",
        "driver_cores",
        "executor_cores",
    ]
    _default_fields_to_strip = KubeResourceSpec._default_fields_to_strip + [
        "driver_node_selector",
        "executor_node_selector",
        "driver_tolerations",
        "executor_tolerations",
        "driver_affinity",
        "executor_affinity",
        "driver_volume_mounts",
        "executor_volume_mounts",
        "driver_cores",
        "executor_cores",
    ]

    __k8s_fields_to_serialize = [
        "driver_volume_mounts",
        "executor_volume_mounts",
        "driver_node_selector",
        "executor_node_selector",
        "executor_affinity",
        "executor_tolerations",
        "driver_affinity",
        "driver_tolerations",
    ]
    _k8s_fields_to_serialize = (
        KubeResourceSpec._k8s_fields_to_serialize + __k8s_fields_to_serialize
    )
    _fields_to_serialize = (
        KubeResourceSpec._fields_to_serialize + __k8s_fields_to_serialize
    )
    _fields_to_skip_validation = KubeResourceSpec._fields_to_skip_validation + [
        # TODO: affinity, tolerations and node_selector are skipped due to preemption mode transitions.
        #  Preemption mode 'none' depends on the previous mode while the default mode may enrich these values.
        #  When we allow 'None' values for these attributes we get their true values and they will undo the default
        #  enrichment when creating the runtime from dict.
        #  The enrichment should move to the server side and then this can be removed.
        "driver_node_selector",
        "executor_node_selector",
        "executor_affinity",
        "executor_tolerations",
        "driver_affinity",
        "driver_tolerations",
    ]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        replicas=None,
        image_pull_policy=None,
        service_account=None,
        image_pull_secret=None,
        driver_resources=None,
        executor_resources=None,
        job_type=None,
        python_version=None,
        spark_version=None,
        restart_policy=None,
        deps=None,
        main_class=None,
        default_handler=None,
        entry_points=None,
        description=None,
        workdir=None,
        build=None,
        spark_conf=None,
        hadoop_conf=None,
        node_selector=None,
        driver_node_selector=None,
        executor_node_selector=None,
        use_default_image=False,
        priority_class_name=None,
        dynamic_allocation=None,
        monitoring=None,
        disable_auto_mount=False,
        pythonpath=None,
        node_name=None,
        affinity=None,
        tolerations=None,
        driver_tolerations=None,
        executor_tolerations=None,
        executor_affinity=None,
        driver_affinity=None,
        preemption_mode=None,
        executor_preemption_mode=None,
        driver_preemption_mode=None,
        driver_volume_mounts=None,
        executor_volume_mounts=None,
        driver_java_options=None,
        executor_java_options=None,
        driver_cores=None,
        executor_cores=None,
        security_context=None,
        state_thresholds=None,
        serving_spec=None,
        graph=None,
        parameters=None,
        track_models=None,
    ):
        super().__init__(
            command=command,
            args=args,
            image=image,
            mode=mode,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            replicas=replicas,
            image_pull_policy=image_pull_policy,
            service_account=service_account,
            image_pull_secret=image_pull_secret,
            default_handler=default_handler,
            entry_points=entry_points,
            description=description,
            workdir=workdir,
            build=build,
            node_selector=node_selector,
            priority_class_name=priority_class_name,
            disable_auto_mount=disable_auto_mount,
            pythonpath=pythonpath,
            node_name=node_name,
            affinity=affinity,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            state_thresholds=state_thresholds,
            serving_spec=serving_spec,
            graph=graph,
            parameters=parameters,
            track_models=track_models,
        )

        self.driver_resources = driver_resources or {}
        self.executor_resources = executor_resources or {}
        self.spark_conf = spark_conf or {}
        self.hadoop_conf = hadoop_conf or {}
        self.job_type = job_type
        self.python_version = python_version
        self.spark_version = spark_version
        self.restart_policy = restart_policy or {}
        self.deps = deps or {}
        self.main_class = main_class
        self.use_default_image = use_default_image
        self.dynamic_allocation = dynamic_allocation or {}
        self.driver_node_selector = driver_node_selector
        self.executor_node_selector = executor_node_selector
        self.monitoring = monitoring or {}
        self.driver_tolerations = driver_tolerations
        self.executor_tolerations = executor_tolerations
        self.executor_affinity = executor_affinity
        self.driver_affinity = driver_affinity
        self.executor_preemption_mode = executor_preemption_mode
        self.driver_preemption_mode = driver_preemption_mode
        self._driver_volume_mounts = {}
        self.driver_volume_mounts = driver_volume_mounts or {}
        self._executor_volume_mounts = {}
        self.executor_volume_mounts = executor_volume_mounts or {}
        self.driver_java_options = driver_java_options
        self.executor_java_options = executor_java_options
        self.driver_cores = driver_cores
        self.executor_cores = executor_cores

    @property
    def executor_tolerations(self) -> list[kubernetes.client.V1Toleration]:
        return self._executor_tolerations

    @executor_tolerations.setter
    def executor_tolerations(self, executor_tolerations):
        self._executor_tolerations = (
            mlrun.runtimes.pod.transform_attribute_to_k8s_class_instance(
                "executor_tolerations", executor_tolerations
            )
        )

    @property
    def driver_tolerations(self) -> list[kubernetes.client.V1Toleration]:
        return self._driver_tolerations

    @driver_tolerations.setter
    def driver_tolerations(self, driver_tolerations):
        self._driver_tolerations = (
            mlrun.runtimes.pod.transform_attribute_to_k8s_class_instance(
                "driver_tolerations", driver_tolerations
            )
        )

    @property
    def executor_affinity(self) -> kubernetes.client.V1Affinity:
        return self._executor_affinity

    @executor_affinity.setter
    def executor_affinity(self, affinity):
        self._executor_affinity = (
            mlrun.runtimes.pod.transform_attribute_to_k8s_class_instance(
                "executor_affinity", affinity
            )
        )

    @property
    def driver_affinity(self) -> kubernetes.client.V1Affinity:
        return self._driver_affinity

    @driver_affinity.setter
    def driver_affinity(self, affinity):
        self._driver_affinity = (
            mlrun.runtimes.pod.transform_attribute_to_k8s_class_instance(
                "executor_affinity", affinity
            )
        )

    @property
    def driver_preemption_mode(self) -> str:
        return self._driver_preemption_mode

    @driver_preemption_mode.setter
    def driver_preemption_mode(self, mode):
        self._driver_preemption_mode = (
            mode or mlrun.mlconf.function_defaults.preemption_mode
        )

    @property
    def executor_preemption_mode(self) -> str:
        return self._executor_preemption_mode

    @executor_preemption_mode.setter
    def executor_preemption_mode(self, mode):
        self._executor_preemption_mode = (
            mode or mlrun.mlconf.function_defaults.preemption_mode
        )

    @property
    def driver_volume_mounts(self) -> list:
        return list(self._driver_volume_mounts.values())

    @driver_volume_mounts.setter
    def driver_volume_mounts(self, volume_mounts):
        self._driver_volume_mounts = {}
        if volume_mounts:
            for volume_mount in volume_mounts:
                self._set_volume_mount(
                    volume_mount, volume_mounts_field_name="_driver_volume_mounts"
                )

    @property
    def executor_volume_mounts(self) -> list:
        return list(self._executor_volume_mounts.values())

    @executor_volume_mounts.setter
    def executor_volume_mounts(self, volume_mounts):
        self._executor_volume_mounts = {}
        if volume_mounts:
            for volume_mount in volume_mounts:
                self._set_volume_mount(
                    volume_mount, volume_mounts_field_name="_executor_volume_mounts"
                )

    def _verify_jvm_memory_string(
        self, resources_field_name: str, memory: typing.Optional[str]
    ):
        if memory:
            verify_field_regex(
                f"function.spec.{resources_field_name}.requests.memory",
                memory,
                [self._jvm_memory_resource_notation],
            )

    def enrich_resources_with_default_pod_resources(
        self, resources_field_name: str, resources: dict
    ):
        if resources_field_name == "driver_resources":
            role = "driver"
        elif resources_field_name == "executor_resources":
            role = "executor"
        else:
            return {}
        resources_types = ["cpu", "memory"]
        resource_requirements = ["requests", "limits"]
        default_resources = mlrun.mlconf.default_spark_resources.to_dict()[role]

        if resources:
            for resource_requirement in resource_requirements:
                for resource_type in resources_types:
                    if (
                        resources.setdefault(resource_requirement, {}).setdefault(
                            resource_type
                        )
                        is None
                    ):
                        resources[resource_requirement][resource_type] = (
                            default_resources[resource_requirement][resource_type]
                        )
        else:
            resources = default_resources

        # Spark operator uses JVM notation for memory, so we must verify it separately
        verify_requests(
            resources_field_name,
            cpu=resources["requests"]["cpu"],
        )
        self._verify_jvm_memory_string(
            resources_field_name, resources["requests"]["memory"]
        )
        resources["requests"] = generate_resources(
            mem=resources["requests"]["memory"], cpu=resources["requests"]["cpu"]
        )
        gpu_type, gpu_value = get_gpu_from_resource_requirement(resources["limits"])
        verify_limits(
            resources_field_name,
            cpu=resources["limits"]["cpu"],
            gpus=gpu_value,
            gpu_type=gpu_type,
        )
        resources["limits"] = generate_resources(
            cpu=resources["limits"]["cpu"],
            gpus=gpu_value,
            gpu_type=gpu_type,
        )
        if not resources["requests"] and not resources["limits"]:
            return {}
        return resources

    def _verify_and_set_requests(
        self,
        resources_field_name,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        patch: bool = False,
    ):
        # Spark operator uses JVM notation for memory, so we must verify it separately
        verify_requests(resources_field_name, cpu=cpu)
        self._verify_jvm_memory_string(resources_field_name, mem)
        resources = generate_resources(mem=mem, cpu=cpu)

        if not patch:
            update_in(
                getattr(self, resources_field_name),
                "requests",
                resources,
            )
        else:
            for resource, resource_value in resources.items():
                update_in(
                    getattr(self, resources_field_name),
                    f"requests.{resource}",
                    resource_value,
                )

    @property
    def driver_resources(self) -> dict:
        return self._driver_resources

    @driver_resources.setter
    def driver_resources(self, resources):
        self._driver_resources = self.enrich_resources_with_default_pod_resources(
            "driver_resources", resources
        )

    @property
    def executor_resources(self) -> dict:
        return self._executor_resources

    @executor_resources.setter
    def executor_resources(self, resources):
        self._executor_resources = self.enrich_resources_with_default_pod_resources(
            "executor_resources", resources
        )


class Spark3Runtime(KubejobRuntime):
    group = "sparkoperator.k8s.io"
    version = "v1beta2"
    apiVersion = group + "/" + version  # noqa: N815
    kind = "spark"
    plural = "sparkapplications"

    # the dot will make the api prefix the configured registry to the image name
    default_mlrun_image = ".spark-job-default-image"
    gpu_suffix = "-cuda"
    code_script = "spark-function-code.py"
    code_path = "/etc/config/mlrun"

    @property
    def spec(self) -> Spark3JobSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", Spark3JobSpec)

    def _get_igz_deps(self):
        return {
            "jars": [
                "local:///spark/v3io-libs/v3io-hcfs_2.12.jar",
                "local:///spark/v3io-libs/v3io-spark3-streaming_2.12.jar",
                "local:///spark/v3io-libs/v3io-spark3-object-dataframe_2.12.jar",
                "local:///igz/java/libs/scala-library-2.12.14.jar",
                "local:///spark/jars/jmx_prometheus_javaagent-0.16.1.jar",
            ],
            "files": ["local:///igz/java/libs/v3io-pyspark.zip"],
        }

    def with_node_selection(
        self,
        node_name: typing.Optional[str] = None,
        node_selector: typing.Optional[dict[str, str]] = None,
        affinity: typing.Optional[kubernetes.client.V1Affinity] = None,
        tolerations: typing.Optional[list[kubernetes.client.V1Toleration]] = None,
    ):
        if node_name:
            raise NotImplementedError(
                "Setting node name is not supported for spark runtime"
            )
        if affinity:
            raise NotImplementedError(
                "Setting affinity is not supported for spark runtime"
            )
        if tolerations:
            raise mlrun.errors.MLRunInvalidArgumentTypeError(
                "Tolerations can be set in spark runtime but not in with_node_selection. "
                "Instead, use with_driver_node_selection and with_executor_node_selection to set tolerations."
            )
        if node_name:
            raise NotImplementedError(
                "Setting node name is not supported for spark runtime"
            )
        mlrun.k8s_utils.validate_node_selectors(node_selector, raise_on_error=False)
        self.with_driver_node_selection(node_name, node_selector, affinity, tolerations)
        self.with_executor_node_selection(
            node_name, node_selector, affinity, tolerations
        )

    def with_driver_node_selection(
        self,
        node_name: typing.Optional[str] = None,
        node_selector: typing.Optional[dict[str, str]] = None,
        affinity: typing.Optional[kubernetes.client.V1Affinity] = None,
        tolerations: typing.Optional[list[kubernetes.client.V1Toleration]] = None,
    ):
        """
        Enables control of which k8s node the spark executor will run on.

        :param node_name:       The name of the k8s node
        :param node_selector:   Label selector, only nodes with matching labels are eligible to be picked
        :param affinity:        Expands the types of constraints you can express - see
                                https://kubernetes.io/docs/concepts/scheduling-eviction/assign-pod-node/#affinity-and-anti-affinity
                                for details
        :param tolerations:     Tolerations are applied to pods, and allow (but do not require) the pods to schedule
                                onto nodes with matching taints - see
                                https://kubernetes.io/docs/concepts/scheduling-eviction/taint-and-toleration
                                for details
        """
        if node_name:
            raise NotImplementedError(
                "Setting node name is not supported for spark runtime"
            )
        if affinity is not None:
            self.spec.driver_affinity = affinity
        if node_selector is not None:
            mlrun.k8s_utils.validate_node_selectors(node_selector, raise_on_error=False)
            self.spec.driver_node_selector = node_selector
        if tolerations is not None:
            self.spec.driver_tolerations = tolerations

    def with_executor_node_selection(
        self,
        node_name: typing.Optional[str] = None,
        node_selector: typing.Optional[dict[str, str]] = None,
        affinity: typing.Optional[kubernetes.client.V1Affinity] = None,
        tolerations: typing.Optional[list[kubernetes.client.V1Toleration]] = None,
    ):
        """
        Enables control of which k8s node the spark executor will run on.

        :param node_name:       The name of the k8s node
        :param node_selector:   Label selector, only nodes with matching labels are eligible to be picked
        :param affinity:        Expands the types of constraints you can express - see
                                https://kubernetes.io/docs/concepts/scheduling-eviction/assign-pod-node/#affinity-and-anti-affinity
                                for details
        :param tolerations:     Tolerations are applied to pods, and allow (but do not require) the pods to schedule
                                onto nodes with matching taints - see
                                https://kubernetes.io/docs/concepts/scheduling-eviction/taint-and-toleration
                                for details
        """
        if node_name:
            raise NotImplementedError(
                "Setting node name is not supported for spark runtime"
            )
        if affinity is not None:
            self.spec.executor_affinity = affinity
        if node_selector is not None:
            mlrun.k8s_utils.validate_node_selectors(node_selector, raise_on_error=False)
            self.spec.executor_node_selector = node_selector
        if tolerations is not None:
            self.spec.executor_tolerations = tolerations

    def with_preemption_mode(
        self, mode: typing.Union[mlrun.common.schemas.function.PreemptionModes, str]
    ):
        """
        Use with_driver_preemption_mode / with_executor_preemption_mode to setup preemption_mode for spark operator
        """
        raise mlrun.errors.MLRunInvalidArgumentTypeError(
            "with_preemption_mode is not supported, use with_driver_preemption_mode / with_executor_preemption_mode"
            " to set preemption mode for spark operator"
        )

    def with_driver_preemption_mode(
        self, mode: typing.Union[mlrun.common.schemas.function.PreemptionModes, str]
    ):
        """
        Preemption mode controls whether the spark driver can be scheduled on preemptible nodes.
        Tolerations, node selector, and affinity are populated on preemptible nodes corresponding to the function spec.

        The supported modes are:

        * **allow** - The function can be scheduled on preemptible nodes
        * **constrain** - The function can only run on preemptible nodes
        * **prevent** - The function cannot be scheduled on preemptible nodes
        * **none** - No preemptible configuration will be applied on the function

        The default preemption mode is configurable in mlrun.mlconf.function_defaults.preemption_mode.
        By default it's set to **prevent**

        :param mode: allow | constrain | prevent | none defined in :py:class:`~mlrun.common.schemas.PreemptionModes`
        """
        preemption_mode = mlrun.common.schemas.function.PreemptionModes(mode)
        self.spec.driver_preemption_mode = preemption_mode.value

    def with_executor_preemption_mode(
        self, mode: typing.Union[mlrun.common.schemas.function.PreemptionModes, str]
    ):
        """
        Preemption mode controls whether the spark executor can be scheduled on preemptible nodes.
        Tolerations, node selector, and affinity are populated on preemptible nodes corresponding to the function spec.

        The supported modes are:

        * **allow** - The function can be scheduled on preemptible nodes
        * **constrain** - The function can only run on preemptible nodes
        * **prevent** - The function cannot be scheduled on preemptible nodes
        * **none** - No preemptible configuration will be applied on the function

        The default preemption mode is configurable in mlrun.mlconf.function_defaults.preemption_mode,
        by default it's set to **prevent**

        :param mode: allow | constrain | prevent | none defined in :py:class:`~mlrun.common.schemas.PreemptionModes`
        """
        preemption_mode = mlrun.common.schemas.function.PreemptionModes(mode)
        self.spec.executor_preemption_mode = preemption_mode.value

    def with_security_context(
        self, security_context: kubernetes.client.V1SecurityContext
    ):
        """
        With security context is not supported for spark runtime.
        Driver / Executor processes run with uid / gid 1000 as long as security context is not defined.
        If in the future we want to support setting security context it will work only from spark version 3.2 onwards.
        """
        raise mlrun.errors.MLRunInvalidArgumentTypeError(
            "with_security_context is not supported with spark operator"
        )

    def with_driver_host_path_volume(
        self,
        host_path: str,
        mount_path: str,
        type: str = "",
        volume_name: str = "host-path-volume",
    ):
        """
        Add a host path volume and mounts it to the driver pod
        More info: https://kubernetes.io/docs/concepts/storage/volumes#hostpath

        :param host_path:   Path of the directory on the host. If the path is a symlink, it follows the link to the
                            real path
        :param mount_path:  Path within the container at which the volume should be mounted.  Must not contain ':'
        :param type:        Type for HostPath Volume Defaults to ""
        :param volume_name: Volume's name. Must be a DNS_LABEL and unique within the pod
        """
        self._with_host_path_volume(
            "_driver_volume_mounts", host_path, mount_path, type, volume_name
        )

    def with_executor_host_path_volume(
        self,
        host_path: str,
        mount_path: str,
        type: str = "",
        volume_name: str = "host-path-volume",
    ):
        """
        Add an host path volume and mount it to the executor pod/s
        More info: https://kubernetes.io/docs/concepts/storage/volumes#hostpath

        :param host_path:   Path of the directory on the host. If the path is a symlink, it follows the link to the
                            real path
        :param mount_path:  Path within the container at which the volume should be mounted.  Must not contain ':'
        :param type:        Type for HostPath Volume Defaults to ""
        :param volume_name: Volume's name. Must be a DNS_LABEL and unique within the pod
        """
        self._with_host_path_volume(
            "_executor_volume_mounts", host_path, mount_path, type, volume_name
        )

    def _with_host_path_volume(
        self,
        volume_mounts_field_name,
        host_path: str,
        mount_path: str,
        type_: str = "",
        volume_name: str = "host-path-volume",
    ):
        volume = kubernetes.client.V1Volume(
            name=volume_name,
            host_path=kubernetes.client.V1HostPathVolumeSource(
                path=host_path, type=type_
            ),
        )
        volume_mount = kubernetes.client.V1VolumeMount(
            mount_path=mount_path, name=volume_name
        )
        kubernetes_api_client = kubernetes.client.ApiClient()
        self.spec.update_vols_and_mounts(
            [kubernetes_api_client.sanitize_for_serialization(volume)],
            [kubernetes_api_client.sanitize_for_serialization(volume_mount)],
            volume_mounts_field_name,
        )

    def with_dynamic_allocation(
        self, min_executors=None, max_executors=None, initial_executors=None
    ):
        """
        Allows to configure spark's dynamic allocation

        :param min_executors:     Min. number of executors
        :param max_executors:     Max. number of executors
        :param initial_executors: Initial number of executors
        """
        self.spec.dynamic_allocation["enabled"] = True
        if min_executors:
            self.spec.dynamic_allocation["minExecutors"] = min_executors
        if max_executors:
            self.spec.dynamic_allocation["maxExecutors"] = max_executors
        if initial_executors:
            self.spec.dynamic_allocation["initialExecutors"] = initial_executors

    def disable_monitoring(self):
        self.spec.monitoring["enabled"] = False

    def _with_monitoring(self, enabled=True, exporter_jar=None):
        self.spec.monitoring["enabled"] = enabled
        if enabled:
            if exporter_jar:
                self.spec.monitoring["exporter_jar"] = exporter_jar

    def with_igz_spark(self, mount_v3io_to_executor=True):
        """
        Configures the pods (driver and executors) to have V3IO access (via file system and via Hadoop).

        :param mount_v3io_to_executor: When False, limits the file system mount to driver pod only. Default is True.
        """
        self._update_igz_jars(deps=self._get_igz_deps())
        self.apply(mount_v3io(name="v3io"))

        # if we only want to mount v3io on the driver, move v3io
        # mounts from common volume mounts to driver volume mounts
        if not mount_v3io_to_executor:
            v3io_mounts = []
            non_v3io_mounts = []
            for mount in self.spec.volume_mounts:
                if get_item_name(mount) == "v3io":
                    v3io_mounts.append(mount)
                else:
                    non_v3io_mounts.append(mount)
            self.spec.volume_mounts = non_v3io_mounts
            self.spec.driver_volume_mounts += v3io_mounts

        self.apply(
            mount_v3iod(
                namespace=config.namespace,
                v3io_config_configmap="spark-operator-v3io-config",
            )
        )
        if config.spark_history_server_path:
            self.spec.spark_conf["spark.eventLog.enabled"] = "true"
            self.spec.spark_conf["spark.eventLog.dir"] = (
                "file://" + config.spark_history_server_path
            )
        if "enabled" not in self.spec.monitoring or self.spec.monitoring["enabled"]:
            self._with_monitoring(
                exporter_jar="/spark/jars/jmx_prometheus_javaagent-0.16.1.jar",
            )

    def with_cores(
        self,
        executor_cores: typing.Optional[int] = None,
        driver_cores: typing.Optional[int] = None,
    ):
        """
        Allows to configure spark.executor.cores and spark.driver.cores parameters. The values must be integers
        greater than or equal to 1. If a parameter is not specified, it defaults to 1.

        Spark operator has multiple options to control the number of cores available to the executor and driver.
        The .coreLimit and .coreRequest parameters can be set for both executor and driver,
        but they only control the k8s properties of the pods created to run the driver/executor.
        Spark itself uses the spec.[executor|driver].cores parameter to set the parallelism of tasks and cores
        assigned to each task within the pod. This function sets the .cores parameters for the job executed.

        See https://github.com/kubeflow/spark-operator/issues/581 for a discussion about those
        parameters and their meaning in Spark operator.

        :param executor_cores: Number of cores to use for executor (spark.executor.cores)
        :param driver_cores:   Number of cores to use for driver (spark.driver.cores)
        """
        if executor_cores:
            if not isinstance(executor_cores, int) or executor_cores < 1:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"executor_cores must be an integer greater than or equal to 1. Got: {executor_cores}"
                )
            self.spec.executor_cores = executor_cores

        if driver_cores:
            if not isinstance(driver_cores, int) or driver_cores < 1:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"driver_cores must be an integer greater than or equal to 1. Got: {driver_cores}"
                )
            self.spec.driver_cores = driver_cores

    @classmethod
    def _get_default_deployed_mlrun_image_name(cls, with_gpu=False):
        suffix = cls.gpu_suffix if with_gpu else ""
        return cls.default_mlrun_image + suffix

    @classmethod
    def deploy_default_image(cls, with_gpu=False):
        if not mlrun.get_current_project(silent=True):
            raise mlrun.errors.MLRunMissingProjectError(
                "An active project is required to run deploy_default_image(). "
                "This can be set by calling get_or_create_project()."
            )

        sj = mlrun.new_function(kind=cls.kind, name="spark-default-image-deploy-temp")
        sj.spec.build.image = cls._get_default_deployed_mlrun_image_name(with_gpu)

        # setting required resources
        sj.with_executor_requests(cpu=1, mem="512m")
        sj.with_driver_requests(cpu=1, mem="512m")

        sj.deploy()
        mlrun.db.get_run_db().delete_function(name=sj.metadata.name)

    def _is_using_gpu(self):
        driver_limits = self.spec.driver_resources.get("limits")
        driver_gpu = None
        if driver_limits:
            _, driver_gpu = self._get_gpu_type_and_quantity(resources=driver_limits)

        executor_limits = self.spec.executor_resources.get("limits")
        executor_gpu = None
        if executor_limits:
            _, executor_gpu = self._get_gpu_type_and_quantity(resources=executor_limits)

        return bool(driver_gpu or executor_gpu)

    @property
    def _default_image(self):
        if config.spark_app_image_tag and config.spark_app_image:
            return (
                config.spark_app_image
                + (self.gpu_suffix if self._is_using_gpu() else "")
                + ":"
                + config.spark_app_image_tag
            )
        return None

    def deploy(
        self,
        watch=True,
        with_mlrun=True,
        skip_deployed=False,
        is_kfp=False,
        mlrun_version_specifier=None,
        builder_env: typing.Optional[dict] = None,
        show_on_failure: bool = False,
        force_build: bool = False,
    ):
        """deploy function, build container with dependencies

        :param watch:                   wait for the deploy to complete (and print build logs)
        :param with_mlrun:              add the current mlrun package to the container build
        :param skip_deployed:           skip the build if we already have an image for the function
        :param is_kfp:                  deploy as part of a kfp pipeline
        :param mlrun_version_specifier: which mlrun package version to include (if not current)
        :param builder_env:             Kaniko builder pod env vars dict (for config/credentials)
                                        e.g. builder_env={"GIT_TOKEN": token}
        :param show_on_failure:         show logs only in case of build failure
        :param force_build:             set True for force building the image, even when no changes were made

        :return: True if the function is ready (deployed)
        """
        # connect will populate the config from the server config
        mlrun.db.get_run_db()
        if not self.spec.build.base_image:
            self.spec.build.base_image = self._default_image
        return super().deploy(
            watch=watch,
            with_mlrun=with_mlrun,
            skip_deployed=skip_deployed,
            is_kfp=is_kfp,
            mlrun_version_specifier=mlrun_version_specifier,
            builder_env=builder_env,
            show_on_failure=show_on_failure,
            force_build=force_build,
        )

    @staticmethod
    def _get_gpu_type_and_quantity(resources):
        gpu_type = [
            resource_type
            for resource_type in resources.keys()
            if resource_type not in ["cpu", "memory"]
        ]
        if len(gpu_type) > 1:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Sparkjob supports only a single gpu type"
            )
        gpu_quantity = resources[gpu_type[0]] if gpu_type else 0
        return gpu_type[0] if gpu_type else None, gpu_quantity

    def _pre_run(self, runobj: RunObject, execution: MLClientCtx):
        if self.spec.build.source and self.spec.build.load_source_on_run:
            raise mlrun.errors.MLRunPreconditionFailedError(
                "Sparkjob does not support loading source code on run, "
                "use func.with_source_archive(pull_at_runtime=False)"
            )

        super()._pre_run(runobj, execution)

    def _update_igz_jars(self, deps):
        if not self.spec.deps:
            self.spec.deps = {}
        if "jars" in deps:
            if "jars" not in self.spec.deps:
                self.spec.deps["jars"] = []
            self.spec.deps["jars"] += deps["jars"]
        if "files" in deps:
            if "files" not in self.spec.deps:
                self.spec.deps["files"] = []
            self.spec.deps["files"] += deps["files"]

    def with_limits(
        self,
        mem=None,
        cpu=None,
        gpus=None,
        gpu_type="nvidia.com/gpu",
        patch: bool = False,
    ):
        raise NotImplementedError(
            "In spark runtimes, use 'with_driver_limits' & 'with_executor_limits'"
        )

    def with_requests(self, mem=None, cpu=None, patch: bool = False):
        raise NotImplementedError(
            "In spark runtimes, use 'with_driver_requests' & 'with_executor_requests'"
        )

    def gpus(self, gpus, gpu_type="nvidia.com/gpu"):
        raise NotImplementedError(
            "In spark runtimes, use 'with_driver_limits' & 'with_executor_limits'"
        )

    def with_executor_requests(
        self,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        patch: bool = False,
    ):
        """
        set executor pod required cpu/memory/gpu resources
        by default it overrides the whole requests section, if you wish to patch specific resources use `patch=True`.
        """
        self.spec._verify_and_set_requests("executor_resources", mem, cpu, patch=patch)

    def with_executor_limits(
        self,
        cpu: typing.Optional[str] = None,
        gpus: typing.Optional[int] = None,
        gpu_type: str = "nvidia.com/gpu",
        patch: bool = False,
    ):
        """
        set executor pod limits
        by default it overrides the whole limits section, if you wish to patch specific resources use `patch=True`.
        """
        # in spark operator there is only use of mem passed through requests,
        # limits is set to the same value so passing mem=None
        self.spec._verify_and_set_limits(
            "executor_resources", None, cpu, gpus, gpu_type, patch=patch
        )

    def with_driver_requests(
        self,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        patch: bool = False,
    ):
        """
        set driver pod required cpu/memory/gpu resources
        by default it overrides the whole requests section, if you wish to patch specific resources use `patch=True`.
        """
        self.spec._verify_and_set_requests("driver_resources", mem, cpu, patch=patch)

    def with_driver_limits(
        self,
        cpu: typing.Optional[str] = None,
        gpus: typing.Optional[int] = None,
        gpu_type: str = "nvidia.com/gpu",
        patch: bool = False,
    ):
        """
        set driver pod cpu limits
        by default it overrides the whole limits section, if you wish to patch specific resources use `patch=True`.
        """
        # in spark operator there is only use of mem passed through requests,
        # limits is set to the same value so passing mem=None
        self.spec._verify_and_set_limits(
            "driver_resources", None, cpu, gpus, gpu_type, patch=patch
        )

    def with_restart_policy(
        self,
        restart_type="OnFailure",
        retries=0,
        retry_interval=10,
        submission_retries=3,
        submission_retry_interval=20,
    ):
        """set restart policy
        restart_type=OnFailure/Never/Always"""
        update_in(self.spec.restart_policy, "type", restart_type)
        update_in(self.spec.restart_policy, "retries", retries)
        update_in(self.spec.restart_policy, "retry_interval", retry_interval)
        update_in(self.spec.restart_policy, "submission_retries", submission_retries)
        update_in(
            self.spec.restart_policy,
            "submission_retry_interval",
            submission_retry_interval,
        )

    def with_source_archive(
        self, source, workdir=None, handler=None, pull_at_runtime=True, target_dir=None
    ):
        """load the code from git/tar/zip archive at runtime or build

        :param source:          valid path to git, zip, or tar file, e.g.
                                git://github.com/mlrun/something.git
                                http://some/url/file.zip
        :param handler:         default function handler
        :param workdir:         working dir relative to the archive root (e.g. './subdir') or absolute to the image root
        :param pull_at_runtime: not supported for spark runtime, must be False
        :param target_dir:      target dir on runtime pod for repo clone / archive extraction
        """
        if pull_at_runtime:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "pull_at_runtime is not supported for spark runtime, use pull_at_runtime=False"
            )

        super().with_source_archive(
            source, workdir, handler, pull_at_runtime, target_dir
        )

    def is_deployed(self):
        if (
            not self.spec.build.source
            and not self.spec.build.commands
            and not self.spec.build.extra
        ):
            return True
        return super().is_deployed()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing
from ast import FunctionDef, parse, unparse
from base64 import b64decode
from typing import Callable, Optional, Union

import mlrun
import mlrun.runtimes.kubejob as kubejob
import mlrun.runtimes.pod as pod
import mlrun.utils.helpers
from mlrun.errors import MLRunInvalidArgumentError
from mlrun.model import HyperParamOptions, RunObject


def get_log_artifacts_code(runobj: RunObject, task_parameters: dict):
    artifact_json_dir = task_parameters.get(
        "artifact_json_dir",
        mlrun.mlconf.function.databricks.artifact_directory_path,
    )
    artifact_json_path = (
        f"{artifact_json_dir}/mlrun_artifact_{runobj.metadata.uid}.json"
    )
    return (
        log_artifacts_code_template.format(f"/dbfs{artifact_json_path}"),
        artifact_json_path,
    )


def replace_log_artifact_function(code: str, log_artifacts_code: str):
    #  user can use a dummy function in oder to avoid edit his code.
    #  replace mlrun_log_artifact function if already exist.
    is_replaced = False
    parsed_code = parse(code)
    for node in parsed_code.body:
        if isinstance(node, FunctionDef) and node.name == "mlrun_log_artifact":
            new_function_ast = parse(log_artifacts_code)
            node.args = new_function_ast.body[0].args
            node.body = new_function_ast.body[0].body
            is_replaced = True
            break
    return unparse(parsed_code), is_replaced


class DatabricksSpec(pod.KubeResourceSpec):
    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        default_handler=None,
        pythonpath=None,
        entry_points=None,
        description=None,
        workdir=None,
        replicas=None,
        image_pull_policy=None,
        service_account=None,
        build=None,
        image_pull_secret=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        disable_auto_mount=False,
        priority_class_name=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        state_thresholds=None,
    ):
        super().__init__(
            command=command,
            image=image,
            mode=mode,
            build=build,
            entry_points=entry_points,
            description=description,
            workdir=workdir,
            default_handler=default_handler,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            replicas=replicas,
            image_pull_policy=image_pull_policy,
            service_account=service_account,
            image_pull_secret=image_pull_secret,
            args=args,
            node_name=node_name,
            node_selector=node_selector,
            affinity=affinity,
            priority_class_name=priority_class_name,
            disable_auto_mount=disable_auto_mount,
            pythonpath=pythonpath,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            state_thresholds=state_thresholds,
        )
        self._termination_grace_period_seconds = 60


class DatabricksRuntime(kubejob.KubejobRuntime):
    kind = "databricks"
    _is_remote = True

    @property
    def spec(self) -> DatabricksSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", DatabricksSpec)

    @staticmethod
    def _verify_returns(returns):
        # TODO complete returns feature
        if returns:
            raise MLRunInvalidArgumentError(
                "Databricks function does not support returns."
            )

    def _get_modified_user_code(self, original_handler: str, log_artifacts_code: str):
        encoded_code: typing.Optional[str] = (
            self.spec.build.functionSourceCode if hasattr(self.spec, "build") else None
        )
        if not encoded_code:
            raise ValueError("Databricks function must be provided with user code")

        decoded_code = b64decode(encoded_code).decode("utf-8")
        decoded_code, is_replaced = replace_log_artifact_function(
            code=decoded_code, log_artifacts_code=log_artifacts_code
        )
        if is_replaced:
            decoded_code = (
                logger_and_consts_code + _databricks_script_code + decoded_code
            )
        else:
            decoded_code = (
                logger_and_consts_code
                + log_artifacts_code
                + _databricks_script_code
                + decoded_code
            )
        if original_handler:
            decoded_code += f"\nresult = {original_handler}(**handler_arguments)\n"
            decoded_code += _return_artifacts_code
        return mlrun.utils.helpers.encode_user_code(decoded_code)

    def get_internal_parameters(self, runobj: RunObject):
        """
        Return the internal function parameters + code.
        """
        task_parameters = runobj.spec.parameters.get("task_parameters", {})
        if "original_handler" in task_parameters:
            original_handler = task_parameters["original_handler"]
        else:
            original_handler = runobj.spec.handler or ""
        log_artifacts_code, artifact_json_path = get_log_artifacts_code(
            runobj=runobj, task_parameters=task_parameters
        )
        returns = runobj.spec.returns or []
        self._verify_returns(returns=returns)
        code = self._get_modified_user_code(
            original_handler=original_handler,
            log_artifacts_code=log_artifacts_code,
        )
        updated_task_parameters = {
            "original_handler": original_handler,
            "artifact_json_path": artifact_json_path,
        }
        return code, updated_task_parameters

    def _pre_run(self, runspec: RunObject, execution):
        internal_code, updated_task_parameters = self.get_internal_parameters(runspec)
        task_parameters = runspec.spec.parameters.get("task_parameters", {})
        task_parameters["spark_app_code"] = internal_code
        for key, value in updated_task_parameters.items():
            if value:
                task_parameters[key] = value  # in order to handle reruns.
        runspec.spec.parameters["task_parameters"] = task_parameters
        wrap_code = b"""
from mlrun.runtimes.databricks_job import databricks_wrapper

def run_mlrun_databricks_job(context,task_parameters: dict, **kwargs):
        databricks_wrapper.run_mlrun_databricks_job(context, task_parameters, **kwargs)
"""
        wrap_code = mlrun.utils.helpers.encode_user_code(wrap_code)
        self.spec.build.functionSourceCode = wrap_code
        runspec.spec.handler = "run_mlrun_databricks_job"

    def run(
        self,
        runspec: Optional[
            Union["mlrun.run.RunTemplate", "mlrun.run.RunObject", dict]
        ] = None,
        handler: Optional[Union[str, Callable]] = None,
        name: Optional[str] = "",
        project: Optional[str] = "",
        params: Optional[dict] = None,
        inputs: Optional[dict[str, str]] = None,
        out_path: Optional[str] = "",
        workdir: Optional[str] = "",
        artifact_path: Optional[str] = "",
        watch: Optional[bool] = True,
        schedule: Optional[Union[str, mlrun.common.schemas.ScheduleCronTrigger]] = None,
        hyperparams: Optional[dict[str, list]] = None,
        hyper_param_options: Optional[HyperParamOptions] = None,
        verbose: Optional[bool] = None,
        scrape_metrics: Optional[bool] = None,
        local: Optional[bool] = False,
        local_code_path: Optional[str] = None,
        auto_build: Optional[bool] = None,
        param_file_secrets: Optional[dict[str, str]] = None,
        notifications: Optional[list[mlrun.model.Notification]] = None,
        returns: Optional[list[Union[str, dict[str, str]]]] = None,
        state_thresholds: Optional[dict[str, int]] = None,
        reset_on_run: Optional[bool] = None,
        output_path: Optional[str] = "",
        retry: Optional[Union[mlrun.model.Retry, dict]] = None,
        **launcher_kwargs,
    ) -> RunObject:
        if local:
            raise MLRunInvalidArgumentError("Databricks runtime cannot run locally.")
        return super().run(
            runspec=runspec,
            handler=handler,
            name=name,
            project=project,
            params=params,
            inputs=inputs,
            out_path=out_path,
            workdir=workdir,
            output_path=output_path,
            watch=watch,
            schedule=schedule,
            hyperparams=hyperparams,
            hyper_param_options=hyper_param_options,
            verbose=verbose,
            scrape_metrics=scrape_metrics,
            local=local,
            local_code_path=local_code_path,
            auto_build=auto_build,
            param_file_secrets=param_file_secrets,
            notifications=notifications,
            returns=returns,
            state_thresholds=state_thresholds,
            **launcher_kwargs,
        )


logger_and_consts_code = """ \n
import os
import logging
mlrun_logger = logging.getLogger('mlrun_logger')
mlrun_logger.setLevel(logging.DEBUG)

mlrun_console_handler = logging.StreamHandler()
mlrun_console_handler.setLevel(logging.DEBUG)
mlrun_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
mlrun_console_handler.setFormatter(mlrun_formatter)
mlrun_logger.addHandler(mlrun_console_handler)

mlrun_default_artifact_template = 'mlrun_return_value_'
mlrun_artifact_index = 0
"""

_databricks_script_code = """

import argparse
import json
parser = argparse.ArgumentParser()
parser.add_argument('handler_arguments')
handler_arguments = parser.parse_args().handler_arguments
handler_arguments = json.loads(handler_arguments)


"""

log_artifacts_code_template = """\n
def mlrun_log_artifact(name='', path=''):
    global mlrun_artifact_index
    mlrun_artifact_index+=1  #  by how many artifacts we tried to log, not how many succeed.
    if name is None or name == '':
        name = f'{{mlrun_default_artifact_template}}{{mlrun_artifact_index}}'
    if not path:
        mlrun_logger.error(f'path required for logging an mlrun artifact - {{name}} : {{path}}')
        return
    if not isinstance(name, str) or not isinstance(path, str):
        mlrun_logger.error(f'name and path must be in string type for logging an mlrun artifact - {{name}} : {{path}}')
        return
    if not path.startswith('/dbfs') and not path.startswith('dbfs:/'):
        mlrun_logger.error(f'path for an mlrun artifact must start with /dbfs or dbfs:/ - {{name}} : {{path}}')
        return
    mlrun_artifacts_path = '{}'
    try:
        new_data = {{name:path}}
        if os.path.exists(mlrun_artifacts_path):
            with open(mlrun_artifacts_path, 'r+') as json_file:
                existing_data = json.load(json_file)
                existing_data.update(new_data)
                json_file.seek(0)
                json.dump(existing_data, json_file)
        else:
            parent_dir = os.path.dirname(mlrun_artifacts_path)
            if parent_dir != '/dbfs':
                os.makedirs(parent_dir, exist_ok=True)
            with open(mlrun_artifacts_path, 'w') as json_file:
                json.dump(new_data, json_file)
        success_log = f'successfully wrote artifact details to the artifact JSON file in DBFS - {{name}} : {{path}}'
        mlrun_logger.info(success_log)
    except Exception as unknown_exception:
        mlrun_logger.error(f'log mlrun artifact failed - {{name}} : {{path}}. error: {{unknown_exception}}')
\n
"""

_return_artifacts_code = """\n
if result:
    if isinstance(result, dict):
        for key, path in result.items():
            mlrun_log_artifact(name=key, path=path)
    elif isinstance(result, (list, tuple, set)):
        for artifact_path in result:
            mlrun_log_artifact(path=artifact_path)
    elif isinstance(result, str):
        mlrun_log_artifact(path=result)
    else:
        mlrun_logger.warning(f'can not log artifacts with the result of handler function \
- result in unsupported type. {type(result)}')
"""



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import json
from abc import ABC, abstractmethod
from typing import Any, Union

import yaml

from ._supported_format import SupportedFormat


class _Formatter(ABC):
    """
    An abstract base class for a formatter - a class to format python structures into and from files.
    """

    @classmethod
    @abstractmethod
    def write(cls, obj: Any, file_path: str, **dump_kwargs: dict):
        """
        Write the object to a file. The object must be serializable according to the used format.

        :param obj:         The object to write.
        :param file_path:   The file path to write to.
        :param dump_kwargs: Additional keyword arguments to pass to the dump method of the formatter in use.
        """
        pass

    @classmethod
    @abstractmethod
    def read(cls, file_path: str) -> Any:
        """
        Read an object from the file given.

        :param file_path: The file to read the object from.

        :return: The read object.
        """
        pass


class _JSONFormatter(_Formatter):
    """
    A static class for managing json files.
    """

    # A set of default configurations to pass to the dump function:
    DEFAULT_DUMP_KWARGS = {"indent": 4}

    @classmethod
    def write(cls, obj: Union[list, dict], file_path: str, **dump_kwargs: dict):
        """
        Write the object to a json file. The object must be serializable according to the json format.

        :param obj:         The object to write.
        :param file_path:   The file path to write to.
        :param dump_kwargs: Additional keyword arguments to pass to the `json.dump` method of the formatter in use.
        """
        dump_kwargs = dump_kwargs or cls.DEFAULT_DUMP_KWARGS
        with open(file_path, "w") as file:
            json.dump(obj, file, **dump_kwargs)

    @classmethod
    def read(cls, file_path: str) -> Union[list, dict]:
        """
        Read an object from the json file given.

        :param file_path: The json file to read the object from.

        :return: The read object.
        """
        with open(file_path) as file:
            obj = json.load(file)
        return obj


class _JSONLFormatter(_Formatter):
    """
    A static class for managing jsonl files.
    """

    @classmethod
    def write(cls, obj: Union[list, dict], file_path: str, **dump_kwargs: dict):
        """
        Write the object to a jsonl file. The object must be serializable according to the json format.

        :param obj:         The object to write.
        :param file_path:   The file path to write to.
        :param dump_kwargs: Additional keyword arguments to pass to the `json.dumps` method of the formatter in use.
        """
        if isinstance(obj, dict):
            obj = [obj]

        with open(file_path, "w") as file:
            for line in obj:
                file.write(json.dumps(obj=line, **dump_kwargs) + "\n")

    @classmethod
    def read(cls, file_path: str) -> Union[list, dict]:
        """
        Read an object from the jsonl file given.

        :param file_path: The jsonl file to read the object from.

        :return: The read object.
        """
        with open(file_path) as file:
            lines = file.readlines()

        obj = []
        for line in lines:
            obj.append(json.loads(s=line))

        return obj[0] if len(obj) == 1 else obj


class _YAMLFormatter(_Formatter):
    """
    A static class for managing yaml files.
    """

    # A set of default configurations to pass to the dump function:
    DEFAULT_DUMP_KWARGS = {"default_flow_style": False, "indent": 4}

    @classmethod
    def write(cls, obj: Union[list, dict], file_path: str, **dump_kwargs: dict):
        """
        Write the object to a yaml file. The object must be serializable according to the yaml format.

        :param obj:         The object to write.
        :param file_path:   The file path to write to.
        :param dump_kwargs: Additional keyword arguments to pass to the `yaml.safe_dump` method of the formatter in use.
        """
        dump_kwargs = dump_kwargs or cls.DEFAULT_DUMP_KWARGS
        with open(file_path, "w") as file:
            yaml.safe_dump(obj, file, **dump_kwargs)

    @classmethod
    def read(cls, file_path: str) -> Union[list, dict]:
        """
        Read an object from the yaml file given.

        :param file_path: The yaml file to read the object from.

        :return: The read object.
        """
        with open(file_path) as file:
            obj = yaml.safe_load(file)
        return obj


class _TXTFormatter(_Formatter):
    """
    A static class for managing txt files.
    """

    @classmethod
    def write(cls, obj: Any, file_path: str, **dump_kwargs: dict):
        """
        Write the object to a text file. The object must be serializable according to python's ast module.

        :param obj:         The object to write.
        :param file_path:   The file path to write to.
        :param dump_kwargs: Ignored.
        """
        with open(file_path, "w") as file:
            file.write(str(obj))

    @classmethod
    def read(cls, file_path: str) -> Any:
        """
        Read an object from the yaml file given.

        :param file_path: The yaml file to read the object from.

        :return: The read object.
        """
        with open(file_path) as file:
            obj = ast.literal_eval(file.read())
        return obj


class StructFileSupportedFormat(SupportedFormat[_Formatter]):
    """
    Library of struct formats (file extensions) supported by some builtin MLRun packagers.
    """

    JSON = "json"
    JSONL = "jsonl"
    YAML = "yaml"
    TXT = "txt"

    _FORMAT_HANDLERS_MAP = {
        JSON: _JSONFormatter,
        JSONL: _JSONLFormatter,
        YAML: _YAMLFormatter,
        TXT: _TXTFormatter,
    }



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from datetime import datetime
from typing import Optional

import fsspec
import v3io
from v3io.dataplane.response import HttpResponseError

import mlrun

from ..platforms.iguazio import parse_path, split_path
from .base import (
    DataStore,
    FileStats,
    basic_auth_header,
)

V3IO_LOCAL_ROOT = "v3io"
V3IO_DEFAULT_UPLOAD_CHUNK_SIZE = 1024 * 1024 * 10


class V3ioStore(DataStore):
    def __init__(
        self, parent, schema, name, endpoint="", secrets: Optional[dict] = None
    ):
        super().__init__(parent, name, schema, endpoint, secrets=secrets)
        self.endpoint = self.endpoint or mlrun.mlconf.v3io_api

        self.headers = None
        self.secure = self.kind == "v3ios"

        token = self._get_secret_or_env("V3IO_ACCESS_KEY")
        username = self._get_secret_or_env("V3IO_USERNAME")
        password = self._get_secret_or_env("V3IO_PASSWORD")
        if self.endpoint.startswith("https://"):
            self.endpoint = self.endpoint[len("https://") :]
            self.secure = True
        elif self.endpoint.startswith("http://"):
            self.endpoint = self.endpoint[len("http://") :]
            self.secure = False
        self.client = v3io.dataplane.Client(access_key=token, endpoint=self.url)
        self.object = self.client.object
        self.auth = None
        self.token = token
        if token:
            self.headers = {"X-v3io-session-key": token}
        elif username and password:
            self.headers = basic_auth_header(username, password)

    @staticmethod
    def _do_object_request(function: callable, *args, **kwargs):
        try:
            return function(*args, **kwargs)
        except HttpResponseError as http_response_error:
            raise mlrun.errors.err_for_status_code(
                status_code=http_response_error.status_code,
                message=mlrun.errors.err_to_str(http_response_error),
            )

    @staticmethod
    def uri_to_ipython(endpoint, subpath):
        return V3IO_LOCAL_ROOT + subpath

    @property
    def url(self):
        schema = "https" if self.secure else "http"
        return f"{schema}://{self.endpoint}"

    @property
    def spark_url(self):
        return "v3io:/"

    @property
    def filesystem(self):
        """return fsspec file system object, if supported"""
        if self._filesystem:
            return self._filesystem
        self._filesystem = fsspec.filesystem("v3io", **self.get_storage_options())
        return self._filesystem

    def get_storage_options(self):
        res = dict(
            v3io_access_key=self._get_secret_or_env("V3IO_ACCESS_KEY"),
            v3io_api=mlrun.mlconf.v3io_api,
        )
        return self._sanitize_options(res)

    def _upload(
        self,
        key: str,
        src_path: str,
        max_chunk_size: int = V3IO_DEFAULT_UPLOAD_CHUNK_SIZE,
    ):
        """helper function for upload method, allows for controlling max_chunk_size in testing"""
        container, path = split_path(self._join(key))
        with open(src_path, "rb") as file_obj:
            append = False
            while True:
                data = memoryview(file_obj.read(max_chunk_size))
                if not data:
                    break
                self._do_object_request(
                    self.object.put,
                    container=container,
                    path=path,
                    body=data,
                    append=append,
                )
                append = True

    def upload(self, key, src_path):
        return self._upload(key, src_path)

    def get(self, key, size=None, offset=0):
        container, path = split_path(self._join(key))
        return self._do_object_request(
            function=self.object.get,
            container=container,
            path=path,
            offset=offset,
            num_bytes=size,
        ).body

    def _put(
        self,
        key,
        data,
        append=False,
        max_chunk_size: int = V3IO_DEFAULT_UPLOAD_CHUNK_SIZE,
    ):
        """helper function for put method, allows for controlling max_chunk_size in testing"""
        data, _ = self._prepare_put_data(data, append)
        container, path = split_path(self._join(key))
        buffer_size = len(data)  # in bytes
        buffer_offset = 0
        try:
            data = memoryview(data)
        except TypeError:
            pass

        while buffer_offset < buffer_size:
            chunk_size = min(buffer_size - buffer_offset, max_chunk_size)
            append = True if buffer_offset or append else False
            self._do_object_request(
                self.object.put,
                container=container,
                path=path,
                body=data[buffer_offset : buffer_offset + chunk_size],
                append=append,
            )
            buffer_offset += chunk_size

    def put(self, key, data, append=False):
        return self._put(key, data, append)

    def stat(self, key):
        container, path = split_path(self._join(key))
        response = self._do_object_request(
            function=self.object.head, container=container, path=path
        )
        head = dict(response.headers)
        size = int(head.get("Content-Length", "0"))
        datestr = head.get("Last-Modified", "0")
        modified = time.mktime(
            datetime.strptime(datestr, "%a, %d %b %Y %H:%M:%S %Z").timetuple()
        )
        return FileStats(size, modified)

    def listdir(self, key):
        container, subpath = split_path(self._join(key))
        if not subpath.endswith("/"):
            subpath += "/"

        # without the trailing slash
        subpath_length = len(subpath) - 1

        try:
            response = self.client.container.list(
                container=container,
                path=subpath,
                get_all_attributes=False,
                directories_only=False,
            )
        except RuntimeError as exc:
            if "Permission denied" in str(exc):
                raise mlrun.errors.MLRunAccessDeniedError(
                    f"Access denied to path: {key}"
                ) from exc
            raise

        # todo: full = key, size, last_modified
        dir_content = [
            dir.prefix[subpath_length:] for dir in response.output.common_prefixes
        ]
        obj_content = [obj.key[subpath_length:] for obj in response.output.contents]
        return dir_content + obj_content

    def rm(self, path, recursive=False, maxdepth=None):
        """Recursive rm file/folder
        Workaround for v3io-fs not supporting recursive directory removal"""

        file_system = self.filesystem
        if isinstance(path, str):
            path = [path]
        maxdepth = maxdepth if not maxdepth else maxdepth - 1
        to_rm = set()
        for p in path:
            _, p = parse_path(p, suffix="")
            p = "/" + p
            if recursive:
                find_out = file_system.find(
                    p, maxdepth=maxdepth, withdirs=True, detail=True
                )
                rec = set(
                    sorted(
                        [
                            f["name"] + ("/" if f["type"] == "directory" else "")
                            for f in find_out.values()
                        ]
                    )
                )
                to_rm |= rec
            if p not in to_rm and (recursive is False or file_system.exists(p)):
                p = p + ("/" if file_system.isdir(p) else "")
                to_rm.add(p)
        for p in reversed(list(sorted(to_rm))):
            file_system.rm_file(p)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1

import mlrun.common.types

from .k8s import Resources


class ProjectMembershipFeatureFlag(mlrun.common.types.StrEnum):
    enabled = "enabled"
    disabled = "disabled"


class PreemptionNodesFeatureFlag(mlrun.common.types.StrEnum):
    enabled = "enabled"
    disabled = "disabled"


class AuthenticationFeatureFlag(mlrun.common.types.StrEnum):
    none = "none"
    basic = "basic"
    bearer = "bearer"
    iguazio = "iguazio"


class NuclioStreamsFeatureFlag(mlrun.common.types.StrEnum):
    enabled = "enabled"
    disabled = "disabled"


class FeatureFlags(pydantic.v1.BaseModel):
    project_membership: ProjectMembershipFeatureFlag
    authentication: AuthenticationFeatureFlag
    nuclio_streams: NuclioStreamsFeatureFlag
    preemption_nodes: PreemptionNodesFeatureFlag


class ArtifactLimits(pydantic.v1.BaseModel):
    max_chunk_size: int
    max_preview_size: int
    max_download_size: int


class FrontendSpec(pydantic.v1.BaseModel):
    jobs_dashboard_url: typing.Optional[str]
    model_monitoring_dashboard_url: typing.Optional[str]
    abortable_function_kinds: list[str] = []
    feature_flags: FeatureFlags
    default_function_priority_class_name: typing.Optional[str]
    valid_function_priority_class_names: list[str] = []
    default_function_image_by_kind: dict[str, str] = {}
    function_deployment_target_image_template: typing.Optional[str]
    function_deployment_target_image_name_prefix_template: str
    function_deployment_target_image_registries_to_enforce_prefix: list[str] = []
    function_deployment_mlrun_requirement: typing.Optional[str]
    auto_mount_type: typing.Optional[str]
    auto_mount_params: dict[str, str] = {}
    default_artifact_path: str
    default_function_pod_resources: Resources = Resources()
    default_function_preemption_mode: str
    feature_store_data_prefixes: typing.Optional[dict[str, str]]
    allowed_artifact_path_prefixes_list: list[str]
    ce: typing.Optional[dict]
    internal_labels: list[str] = []
    artifact_limits: ArtifactLimits



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from os import environ
from typing import Optional

import numpy as np
import pytz
from pyspark.sql.functions import to_utc_timestamp
from pyspark.sql.types import BooleanType, DoubleType

from mlrun.feature_store.retrieval.spark_merger import spark_df_to_pandas
from mlrun.utils import logger

from .data_types import InferOptions, spark_to_value_type

try:
    import pyspark.sql.functions as funcs
except ImportError:
    pass


def infer_schema_from_df_spark(
    df,
    features,
    entities,
    timestamp_key: Optional[str] = None,
    entity_columns=None,
    options: InferOptions = InferOptions.Null,
):
    timestamp_fields = []
    current_entities = list(entities.keys())
    entity_columns = entity_columns or []

    def upsert_entity(name, value_type):
        if name in current_entities:
            old_type = entities[name].value_type
            if old_type != value_type:
                logger.warning(
                    f"Overriding type of entity '{name}' from '{old_type}' to '{value_type}'."
                    " This may result in errors or unusable data."
                )
            entities[name].value_type = value_type
        else:
            entities[name] = {"name": name, "value_type": value_type}

    for column, s in df.dtypes:
        value_type = spark_to_value_type(s)
        is_entity = column in entity_columns or column in current_entities
        if is_entity:
            upsert_entity(column, value_type)
        elif options & InferOptions.Features and column != timestamp_key:
            if column in features.keys():
                features[column].value_type = value_type
            else:
                features[column] = {"name": column, "value_type": value_type}
        if value_type == "timestamp" and not is_entity:
            timestamp_fields.append(column)

    return timestamp_key


def get_df_preview_spark(df, preview_lines=20):
    """capture preview data from spark df"""
    df = df.limit(preview_lines)

    result_dict = spark_df_to_pandas(df).to_dict(orient="split")
    return [result_dict["columns"], *result_dict["data"]]


def _create_hist_data(df, calculation_cols, results_dict, bins=20):
    hist_df = df
    col_names = []
    bins_values = {}

    for orig_col in calculation_cols:
        bins_start = np.linspace(
            float(results_dict[orig_col]["min"]),
            float(results_dict[orig_col]["max"]),
            bins,
            endpoint=False,
        ).tolist()
        bins_values[orig_col] = bins_start

        for i in range(bins):
            col_name = f"{orig_col}_hist_bin_{i}"
            if i < (bins - 1):
                hist_df = hist_df.withColumn(
                    col_name,
                    funcs.when(
                        (funcs.col(orig_col) >= bins_start[i])
                        & (funcs.col(orig_col) < bins_start[i + 1]),
                        1,
                    ).otherwise(0),
                )
            else:
                hist_df = hist_df.withColumn(
                    col_name,
                    funcs.when(funcs.col(orig_col) >= bins_start[i], 1).otherwise(0),
                )
            col_names.append(col_name)

    agg_funcs = [funcs.sum(hist_df[col]).alias(col) for col in col_names]
    hist_values = hist_df.groupBy().agg(*agg_funcs).toPandas()

    for orig_col in calculation_cols:
        bin_data = [
            [int(hist_values.loc[0][f"{orig_col}_hist_bin_{i}"]) for i in range(bins)],
            [round(x, 2) for x in bins_values[orig_col]],
        ]
        results_dict[orig_col]["hist"] = bin_data


def get_dtype(df, colname):
    return [dtype for name, dtype in df.dtypes if name == colname][0]


# Histogram calculation in Spark relies on adding a column to the DF per histogram bin. Spark supports many
# columns, but not infinitely (and there's no hard-limit on number of columns). This value will determine
# how many histograms will be calculated in a single query. By default we're using 20 bins per histogram, so
# using 500 will calculate histograms over 25 columns in a single query. If there are more, more queries will
# be executed.
MAX_HISTOGRAM_COLUMNS_IN_QUERY = int(
    environ.get("MLRUN_MAX_HISTOGRAM_COLUMNS_IN_QUERY", 500)
)


def get_df_stats_spark(df, options, num_bins=20, sample_size=None):
    # todo: sample spark DF if sample_size is not None and DF is bigger than sample_size

    df_after_type_casts = df
    timestamp_columns = set()
    boolean_columns = set()
    for field in df_after_type_casts.schema.fields:
        # covers TimestampType and TimestampNTZType, which was added in PySpark 3.4.0
        is_timestamp = field.dataType.typeName().startswith("timestamp")
        is_boolean = isinstance(field.dataType, BooleanType)
        if is_timestamp:
            df_after_type_casts = df_after_type_casts.withColumn(
                field.name, to_utc_timestamp(df_after_type_casts[field.name], "UTC")
            )
            timestamp_columns.add(field.name)
        if is_boolean:
            boolean_columns.add(field.name)
        if is_timestamp or is_boolean:
            df_after_type_casts = df_after_type_casts.withColumn(
                field.name,
                df_after_type_casts[field.name].cast(DoubleType()),
            )

    # if a column named "summary" already exists, we have to rename it to something else and back
    summary_renamed = False
    df_for_summary = df_after_type_casts
    if "summary" in df.columns:
        df_for_summary = df.withColumnRenamed("summary", "__summary_internal__")
        summary_renamed = True

    summary_df = df_for_summary.summary().toPandas()
    summary_df.set_index(["summary"], drop=True, inplace=True)
    if summary_renamed:
        summary_df.rename(columns={"__summary_internal__": "summary"}, inplace=True)
    # pandas df.describe() returns std, while spark returns stddev
    # we therefore need to rename stddev to std for compatibility with pandas
    # TODO: we may want to consider changing std to stddev in 1.2 and beyond (this requires a change to mlrun-ui)
    summary_df.rename(index={"stddev": "std"}, inplace=True)

    results_dict = {}
    hist_columns = []
    # Spark summary() returns strings, unlike pandas describe() which returns numerical values where
    # applicable. For compatibility, we therefore convert values to numerical types in these cases.
    numerical_spark_types = {"int", "bigint", "float", "double", "bigdecimal"}
    for col, values in summary_df.items():
        original_type = None
        for col_name, col_type in df_after_type_casts.dtypes:
            if col_name == col:
                original_type = col_type
                break
        stats_dict = {}
        for stat, val in values.dropna().items():
            if stat in ["min", "max"] and original_type not in numerical_spark_types:
                stats_dict[stat] = val
            else:
                stats_dict[stat] = float(val)
        results_dict[col] = stats_dict

        if (
            InferOptions.get_common_options(options, InferOptions.Histogram)
            and get_dtype(df_after_type_casts, col) in ["double", "bigint", "int"]
            # in some situations (empty DF may cause this) we won't have stats for columns with suitable types, in this
            # case we won't calculate histogram for the column.
            and "min" in results_dict[col]
            and "max" in results_dict[col]
        ):
            hist_columns.append(col)

    # We may need multiple queries here. See comment before this function for reasoning.
    max_columns_per_query = int(MAX_HISTOGRAM_COLUMNS_IN_QUERY // num_bins) or 1
    while len(hist_columns) > 0:
        calculation_cols = hist_columns[:max_columns_per_query]
        _create_hist_data(df_after_type_casts, calculation_cols, results_dict, num_bins)
        hist_columns = hist_columns[max_columns_per_query:]

    # convert values back to timestamp and boolean where appropriate
    original_type_stats = {"min", "max", "25%", "50%", "75%"}
    for col, stats in results_dict.items():
        if col in timestamp_columns:
            for stat, val in stats.items():
                if stat == "mean" or stat in original_type_stats:
                    stats[stat] = datetime.fromtimestamp(val, tz=pytz.UTC).isoformat()
                elif stat == "hist":
                    values = stats[stat][1]
                    for i in range(len(values)):
                        values[i] = datetime.fromtimestamp(
                            values[i], tz=pytz.UTC
                        ).isoformat()
        # for boolean values, keep mean and histogram values numeric (0 to 1 representation)
        if col in boolean_columns:
            for stat, val in stats.items():
                if stat in original_type_stats:
                    stats[stat] = bool(val)

    return results_dict


class SparkDataInfer:
    infer_schema = infer_schema_from_df_spark
    get_preview = get_df_preview_spark
    get_stats = get_df_stats_spark



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import math
import tarfile
import tempfile
import typing
from urllib.parse import parse_qs, urlparse

import pandas as pd
import semver

import mlrun.datastore


def parse_kafka_url(
    url: str, brokers: typing.Optional[typing.Union[list, str]] = None
) -> tuple[str, list]:
    """Generating Kafka topic and adjusting a list of bootstrap servers.

    :param url:               URL path to parse using urllib.parse.urlparse.
    :param brokers: List of kafka brokers.

    :return: A tuple of:
         [0] = Kafka topic value
         [1] = List of bootstrap servers
    """
    brokers = brokers or []

    if isinstance(brokers, str):
        brokers = brokers.split(",")

    # Parse the provided URL into six components according to the general structure of a URL
    url = urlparse(url)

    # Add the network location to the bootstrap servers list
    if url.netloc:
        brokers = [url.netloc] + brokers

    # Get the topic value from the parsed url
    query_dict = parse_qs(url.query)
    if "topic" in query_dict:
        topic = query_dict["topic"][0]
    else:
        topic = url.path
        topic = topic.lstrip("/")
    return topic, brokers


def upload_tarball(source_dir, target, secrets=None):
    # will delete the temp file
    with tempfile.NamedTemporaryFile(suffix=".tar.gz") as temp_fh:
        with tarfile.open(mode="w:gz", fileobj=temp_fh) as tar:
            tar.add(source_dir, arcname="")
        stores = mlrun.datastore.store_manager.set(secrets)
        datastore, subpath, url = stores.get_or_create_store(target)
        datastore.upload(subpath, temp_fh.name)


def filter_df_start_end_time(
    df: typing.Union[pd.DataFrame, typing.Iterator[pd.DataFrame]],
    time_column: typing.Optional[str] = None,
    start_time: pd.Timestamp = None,
    end_time: pd.Timestamp = None,
) -> typing.Union[pd.DataFrame, typing.Iterator[pd.DataFrame]]:
    if not time_column:
        return df
    if isinstance(df, pd.DataFrame):
        return _execute_time_filter(df, time_column, start_time, end_time)
    else:
        return filter_df_generator(df, time_column, start_time, end_time)


def filter_df_generator(
    dfs: typing.Iterator[pd.DataFrame],
    time_field: str,
    start_time: pd.Timestamp,
    end_time: pd.Timestamp,
) -> typing.Iterator[pd.DataFrame]:
    for df in dfs:
        yield _execute_time_filter(df, time_field, start_time, end_time)


def _execute_time_filter(
    df: pd.DataFrame, time_column: str, start_time: pd.Timestamp, end_time: pd.Timestamp
):
    if semver.parse(pd.__version__)["major"] >= 2:
        # pandas 2 is too strict by default (ML-5629)
        kwargs = {
            "format": "mixed",
            "yearfirst": True,
        }
    else:
        # pandas 1 may fail on format "mixed" (ML-5661)
        kwargs = {}
    df[time_column] = pd.to_datetime(df[time_column], **kwargs)
    if start_time:
        df = df[df[time_column] > start_time]
    if end_time:
        df = df[df[time_column] <= end_time]
    return df


def select_columns_from_df(
    df: typing.Union[pd.DataFrame, typing.Iterator[pd.DataFrame]],
    columns: list[str],
) -> typing.Union[pd.DataFrame, typing.Iterator[pd.DataFrame]]:
    if not columns:
        return df
    if isinstance(df, pd.DataFrame):
        return df[columns]
    else:
        return select_columns_generator(df, columns)


def select_columns_generator(
    dfs: typing.Union[pd.DataFrame, typing.Iterator[pd.DataFrame]],
    columns: list[str],
) -> typing.Iterator[pd.DataFrame]:
    for df in dfs:
        yield df[columns]


def _generate_sql_query_with_time_filter(
    table_name: str,
    engine: "sqlalchemy.engine.Engine",  # noqa: F821,
    time_column: str,
    parse_dates: list[str],
    start_time: pd.Timestamp,
    end_time: pd.Timestamp,
):
    # Validate sqlalchemy (not installed by default):
    try:
        import sqlalchemy
    except (ModuleNotFoundError, ImportError) as exc:
        raise mlrun.errors.MLRunMissingDependencyError(
            "Using 'SQLTarget' requires sqlalchemy package. Use pip install mlrun[sqlalchemy] to install it."
        ) from exc
    table = sqlalchemy.Table(
        table_name,
        sqlalchemy.MetaData(),
        autoload_with=engine,
    )
    query = sqlalchemy.select(table)
    if time_column:
        if parse_dates and time_column not in parse_dates:
            parse_dates.append(time_column)
        else:
            parse_dates = [time_column]
        if start_time:
            query = query.filter(getattr(table.c, time_column) > start_time)
        if end_time:
            query = query.filter(getattr(table.c, time_column) <= end_time)

    return query, parse_dates


def get_kafka_brokers_from_dict(options: dict, pop=False) -> typing.Optional[str]:
    get_or_pop = options.pop if pop else options.get
    kafka_brokers = get_or_pop("kafka_brokers", None)
    return kafka_brokers


def transform_list_filters_to_tuple(additional_filters):
    tuple_filters = []
    if not additional_filters:
        return tuple_filters
    validate_additional_filters(additional_filters)
    for additional_filter in additional_filters:
        tuple_filters.append(tuple(additional_filter))
    return tuple_filters


def validate_additional_filters(additional_filters):
    nan_error_message = "using NaN in additional_filters is not supported"
    if additional_filters in [None, [], ()]:
        return
    for filter_tuple in additional_filters:
        if filter_tuple == () or filter_tuple == []:
            continue
        if not isinstance(filter_tuple, (list, tuple)):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"mlrun supports additional_filters only as a list of tuples."
                f" Current additional_filters: {additional_filters}"
            )
        if isinstance(filter_tuple[0], (list, tuple)):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"additional_filters does not support nested list inside filter tuples except in -in- logic."
                f" Current filter_tuple: {filter_tuple}."
            )
        if len(filter_tuple) != 3:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"illegal filter tuple length, {filter_tuple} in additional filters:"
                f" {additional_filters}"
            )
        col_name, op, value = filter_tuple
        if isinstance(value, float) and math.isnan(value):
            raise mlrun.errors.MLRunInvalidArgumentError(nan_error_message)
        elif isinstance(value, (list, tuple)):
            for sub_value in value:
                if isinstance(sub_value, float) and math.isnan(sub_value):
                    raise mlrun.errors.MLRunInvalidArgumentError(nan_error_message)


class KafkaParameters:
    def __init__(self, kwargs: typing.Optional[dict] = None):
        import kafka

        if kwargs is None:
            kwargs = {}
        self._kafka = kafka
        self._kwargs = kwargs
        self._client_configs = {
            "consumer": self._kafka.KafkaConsumer.DEFAULT_CONFIG,
            "producer": self._kafka.KafkaProducer.DEFAULT_CONFIG,
            "admin": self._kafka.KafkaAdminClient.DEFAULT_CONFIG,
        }
        self._custom_attributes = {
            "max_workers": "",
            "brokers": "",
            "topics": "",
            "group": "",
            "initial_offset": "",
            "partitions": "",
            "sasl": "",
            "worker_allocation_mode": "",
            # for Nuclio with Confluent Kafka
            "tls_enable": "",
            "tls": "",
            "new_topic": "",
            "nuclio_annotations": "",
        }
        self._reference_dicts = (
            self._custom_attributes,
            self._kafka.KafkaAdminClient.DEFAULT_CONFIG,
            self._kafka.KafkaProducer.DEFAULT_CONFIG,
            self._kafka.KafkaConsumer.DEFAULT_CONFIG,
        )

        self._validate_keys()

    def _validate_keys(self) -> None:
        for key in self._kwargs:
            if all(key not in d for d in self._reference_dicts):
                raise ValueError(
                    f"Key '{key}' not found in any of the Kafka reference dictionaries"
                )

    def _get_config(self, client_type: str) -> dict:
        res = {
            k: self._kwargs[k]
            for k in self._kwargs.keys() & self._client_configs[client_type].keys()
        }
        if sasl := self._kwargs.get("sasl"):
            res |= {
                "security_protocol": self._kwargs.get(
                    "security_protocol", "SASL_PLAINTEXT"
                ),
                "sasl_mechanism": sasl["mechanism"],
                "sasl_plain_username": sasl["user"],
                "sasl_plain_password": sasl["password"],
            }
        return res

    def consumer(self) -> dict:
        return self._get_config("consumer")

    def producer(self) -> dict:
        return self._get_config("producer")

    def admin(self) -> dict:
        return self._get_config("admin")

    def sasl(
        self, *, usr: typing.Optional[str] = None, pwd: typing.Optional[str] = None
    ) -> dict[str, typing.Union[str, bool]]:
        res = self._kwargs.get("sasl", {})
        usr = usr or self._kwargs.get("sasl_plain_username")
        pwd = pwd or self._kwargs.get("sasl_plain_password")
        if usr and pwd:
            res["enable"] = True
            res["user"] = usr
            res["password"] = pwd
            res["mechanism"] = self._kwargs.get("sasl_mechanism", "PLAIN")
            res["handshake"] = self._kwargs.get("sasl_handshake", True)
        return res

    def tls(self, *, tls_enable: typing.Optional[bool] = None) -> dict[str, bool]:
        res = self._kwargs.get("tls", {})
        tls_enable = (
            tls_enable if tls_enable is not None else self._kwargs.get("tls_enable")
        )
        if tls_enable:
            res["enable"] = tls_enable
        return res

    def valid_entries_only(self, input_dict: dict) -> dict:
        valid_keys = set()
        for ref_dict in self._reference_dicts:
            valid_keys.update(ref_dict.keys())
        # Return a new dictionary with only valid keys
        return {k: v for k, v in input_dict.items() if k in valid_keys}


def parse_url(url):
    if url and url.startswith("v3io://") and not url.startswith("v3io:///"):
        url = url.replace("v3io://", "v3io:///", 1)
    parsed_url = urlparse(url)
    schema = parsed_url.scheme.lower()
    endpoint = parsed_url.hostname

    # Special handling for WASBS URLs to preserve container information
    if schema in ["wasbs", "wasb"] and parsed_url.netloc and "@" in parsed_url.netloc:
        # For wasbs://container@host format, preserve the full netloc as endpoint
        # This allows the datastore to extract container later
        endpoint = parsed_url.netloc
    elif endpoint:
        # HACK - urlparse returns the hostname after in lower case - we want the original case:
        # the hostname is a substring of the netloc, in which it's the original case, so we find the indexes of the
        # hostname in the netloc and take it from there
        lower_hostname = parsed_url.hostname
        netloc = str(parsed_url.netloc)
        lower_netloc = netloc.lower()
        hostname_index_in_netloc = lower_netloc.index(str(lower_hostname))
        endpoint = netloc[
            hostname_index_in_netloc : hostname_index_in_netloc + len(lower_hostname)
        ]
        if parsed_url.port:
            endpoint += f":{parsed_url.port}"
    return schema, endpoint, parsed_url


def accepts_param(func: callable, param_name):
    sig = inspect.signature(func)
    return param_name in sig.parameters



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
import re
import typing
from os.path import exists, isdir
from urllib.parse import urlparse

import mlrun.artifacts.helpers
import mlrun.common.schemas.artifact
import mlrun.config
import mlrun.utils.regex
from mlrun.utils.helpers import (
    get_local_file_schema,
    template_artifact_path,
    validate_inline_artifact_body_size,
)

from ..utils import (
    is_relative_path,
    logger,
    validate_artifact_key_name,
    validate_tag_name,
)
from .base import (
    Artifact,
    DirArtifact,
    LinkArtifact,
)
from .dataset import (
    DatasetArtifact,
    TableArtifact,
)
from .document import DocumentArtifact
from .llm_prompt import LLMPromptArtifact
from .model import ModelArtifact
from .plots import (
    PlotArtifact,
    PlotlyArtifact,
)

artifact_types = {
    "": Artifact,
    "artifact": Artifact,
    "dir": DirArtifact,
    "link": LinkArtifact,
    "plot": PlotArtifact,
    "table": TableArtifact,
    "model": ModelArtifact,
    "dataset": DatasetArtifact,
    "plotly": PlotlyArtifact,
    "document": DocumentArtifact,
    "llm-prompt": LLMPromptArtifact,
}


class ArtifactProducer:
    def __init__(self, kind, project, name, tag=None, owner=None):
        self.kind = kind
        self.project = project
        self.name = name
        self.tag = tag
        self.owner = owner
        self.uri = "/"
        self.iteration = 0
        self.inputs = {}

    def get_meta(self) -> dict:
        return {
            "kind": self.kind,
            "name": self.name,
            "tag": self.tag,
            "owner": self.owner,
        }

    @property
    def uid(self):
        return None

    @staticmethod
    def parse_uri(uri: str) -> tuple[str, str, str]:
        """Parse artifact producer's uri

        :param uri: artifact producer's uri in the format <project>/<uid>[-<iteration>]
        :returns: tuple of project, uid, iteration
        """
        uri_pattern = mlrun.utils.regex.artifact_producer_uri_pattern
        match = re.match(uri_pattern, uri)
        if not match:
            return "", "", ""
        group_dict = match.groupdict()

        return (
            group_dict["project"] or "",
            group_dict["uid"] or "",
            group_dict["iteration"] or "",
        )


def dict_to_artifact(struct: dict) -> Artifact:
    kind = struct.get("kind", "")
    artifact_class = artifact_types[kind]
    return artifact_class.from_dict(struct)


class ArtifactManager:
    def __init__(
        self,
        db: "mlrun.db.RunDBInterface" = None,
        calc_hash=True,
    ):
        self.calc_hash = calc_hash

        self.artifact_db = db
        self.input_artifacts = {}
        self.artifact_uris = {}

    @staticmethod
    def ensure_artifact_source_file_exists(item, path, body):
        # If the body exists, the source path does not have to exists.
        if body is not None or item.get_body() is not None:
            return
        if not path:
            return
        #  ModelArtifact is a directory.
        if isinstance(item, ModelArtifact):
            return
        # Could happen in the import artifact scenario - that path is None.
        if item.target_path:
            return
        #  in DatasetArtifact
        if hasattr(item, "df") and item.df is not None:
            return
        parsed_url = urlparse(path)
        schema = parsed_url.scheme
        #  we are not checking remote paths yet.
        if schema and schema not in get_local_file_schema():
            return
        if schema.lower() == "file":
            path = parsed_url.path
        if not exists(path):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Failed to log an artifact, file does not exists at path {path}"
            )

    def artifact_list(self, full=False):
        artifacts = []
        for artifacts_uri in self.artifact_uris.values():
            artifact: Artifact = mlrun.datastore.get_store_resource(artifacts_uri)
            if full:
                artifacts.append(artifact.to_dict())
            else:
                artifacts.append(artifact.base_dict())
        return artifacts

    def log_artifact(
        self,
        producer: typing.Union["ArtifactProducer", "mlrun.MLClientCtx"],
        item: Artifact,
        body=None,
        target_path="",
        tag="",
        viewer="",
        local_path="",
        artifact_path=None,
        format=None,
        upload=None,
        labels=None,
        db_key=None,
        project=None,
        is_retained_producer=None,
        **kwargs,
    ) -> Artifact:
        """
        Log an artifact to the DB and upload it to the artifact store.
        :param producer: The producer of the artifact, the producer depends on where the artifact is being logged.
        :param item: The artifact to log.
        :param body: The body of the artifact.
        :param target_path: The target path of the artifact. (cannot be a relative path)
                            If not provided, the artifact will be stored in the default artifact path.
                            If provided and is a remote path (e.g. s3://bucket/path), no artifact will be uploaded
                            as it already exists.
        :param tag: The tag of the artifact.
        :param viewer: Kubeflow viewer type
        :param local_path: The local path of the artifact. If remote path is provided then the artifact won't be
                            uploaded and this parameter will be used as the target path.
        :param artifact_path: The path to store the artifact.
         If not provided, the artifact will be stored in the default artifact path.
        :param format: The format of the artifact. (e.g. csv, json, html, etc.)
        :param upload: Whether to upload the artifact to the datastore. If not provided, and the
        `local_path` is not a directory, upload occurs by default. Directories are uploaded only when this
        flag is explicitly set to `True`.
        :param labels: Labels to add to the artifact.
        :param db_key: The key to use when logging the artifact to the DB.
        If not provided, will generate a key based on the producer name and the artifact key.
        :param project: The project to log the artifact to. If not provided, will use the producer's project.
        :param is_retained_producer: Whether the producer is retained or not. Relevant to register artifacts flow
        where a project may log artifacts which were produced by another producer.
        :param kwargs: Arguments to pass to the artifact class.
        :return: The logged artifact.
        """
        if isinstance(item, str):
            key = item
            if local_path and isdir(local_path):
                item = DirArtifact(key, body, **kwargs)
            else:
                item = Artifact(key, body, **kwargs)
        else:
            key = item.key
            target_path = target_path or item.target_path
        if isinstance(item, ModelArtifact) and item.model_url:
            if upload:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "log_artifact of ModelArtifact does not accept arguments for both upload and model_url parameters"
                )
            upload = False
        validate_artifact_key_name(key, "artifact.key")
        validate_inline_artifact_body_size(item.spec.inline)
        src_path = local_path or item.src_path  # TODO: remove src_path
        self.ensure_artifact_source_file_exists(item=item, path=src_path, body=body)
        if format == "html" or (src_path and pathlib.Path(src_path).suffix == "html"):
            viewer = "web-app"
        item.format = format or item.format
        item.src_path = src_path

        if db_key is None:
            # set the default artifact db key
            if producer.kind == "run" and not is_retained_producer:
                # When the producer's type is "run,"
                # we generate a different db_key than the one we obtained in the request.
                # As a result, a new artifact for the requested key will be created,
                # which will contain the new db_key and will represent the current run.
                # We implement this so that the user can query an artifact,
                # and receive back all the runs that are associated with his search result.
                db_key = producer.name + "_" + key
            else:
                # if the db_key is not explicitly set on the item, we want to use the key as the db_key
                # otherwise, we do not want to override it.
                # this is mainly relevant for imported artifacts that have an explicit db_key value already set
                db_key = item.db_key or key
        if db_key != key:
            validate_artifact_key_name(db_key, "artifact.db_key")
        item.db_key = db_key or ""
        item.viewer = viewer or item.viewer
        item.tree = producer.tag
        item.tag = tag or item.tag

        item.producer = producer.get_meta()
        item.labels = labels or item.labels
        # if running as part of a workflow, enrich artifact with workflow uid label
        if item.producer.get("workflow"):
            item.labels.update({"workflow-id": item.producer.get("workflow")})

        item.iter = producer.iteration
        project = project or producer.project
        item.project = project
        if item.spec.parent_uri:
            mlrun.artifacts.helpers.check_artifact_parent(
                artifact_project=item.project, expected_parent_uri=item.spec.parent_uri
            )
        if is_retained_producer:
            # if the producer is retained, we want to use the original target path
            target_path = target_path or item.target_path

        # if target_path is provided and not relative, then no need to upload the artifact as it already exists
        if target_path:
            if is_relative_path(target_path):
                raise ValueError(
                    f"target_path ({target_path}) param cannot be relative"
                )
            if upload is None:
                upload = False

        # if target_path wasn't provided, but src_path is not relative, then no need to upload the artifact as it
        # already exists. In this case set the target_path to the src_path and set upload to False
        elif src_path and "://" in src_path:
            if upload:
                raise ValueError(f"Cannot upload from remote path {src_path}")
            target_path = src_path
            upload = False

        # if mlrun.mlconf.generate_target_path_from_artifact_hash outputs True and the user
        # didn't pass target_path explicitly then we won't use `generate_target_path` to calculate the target path,
        # but rather use the `resolve_<body/file>_target_hash_path` in the `item.upload` method.
        elif (
            # if the user didn't pass target_path explicitly but asked for upload (or didn't set upload at all)
            # and the other conditions match we will enrich the target_path.
            # generally we don't want to enrich target_path if there is no matching artifact in target_path either
            # there is already a remote artifact in target_path ( that was previously uploaded ) or the user asked
            # to upload the artifact.
            (upload or upload is None)
            and not item.is_inline()
            and not mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash
        ):
            target_path = item.generate_target_path(artifact_path, producer)

        if target_path and item.is_dir and not target_path.endswith("/"):
            target_path += "/"
        target_path = template_artifact_path(
            artifact_path=target_path, project=producer.project, run_uid=producer.uid
        )
        item.target_path = target_path

        item.before_log()

        if ((upload is None and item.kind != "dir") or upload) and not item.is_inline():
            # before uploading the item, we want to ensure that its tags are valid,
            # so that we don't upload something that won't be stored later
            validate_tag_name(item.metadata.tag, "artifact.metadata.tag")
            item.upload(artifact_path=artifact_path)

        if db_key:
            artifact_uid = self._log_to_db(db_key, project, producer.inputs, item)
            if artifact_uid is not None:
                item.uid = artifact_uid
        # Generate the artifact URI after logging to the database and retrieving the artifact UID, if available.
        self.artifact_uris[key] = item.uri

        size = str(item.size) or "?"
        db_str = "Y" if (self.artifact_db and db_key) else "N"
        logger.debug(
            f"Log artifact {key} at {item.target_path}, size: {size}, db: {db_str}"
        )
        return item

    def update_artifact(self, producer, item: Artifact):
        self.artifact_uris[item.key] = item.uri
        self._log_to_db(item.db_key, producer.project, producer.inputs, item)

    def _log_to_db(self, key, project, sources, item, tag=None) -> typing.Optional[str]:
        """
        log artifact to db
        :param key: Identifying key of the artifact.
        :param project: Project that the artifact belongs to.
        :param sources: List of artifact sources ( Mainly passed from the `producer.items` ).
        :param item: The actual artifact to store.
        :param tag: The name of the Tag of the artifact.
        :return: The logged artifact uid.
        """
        if self.artifact_db:
            item.updated = None
            if sources:
                item.sources = [{"name": k, "path": str(v)} for k, v in sources.items()]
            artifact_item = self.artifact_db.store_artifact(
                key,
                item.to_dict(),
                iter=item.iter,
                tag=tag or item.tag,
                project=project,
                tree=item.tree,
            )
            if artifact_item:
                return artifact_item.get("metadata", {}).get("uid")

    def link_artifact(
        self,
        project,
        name,
        tree,
        key,
        iter=0,
        artifact_path="",
        tag="",
        link_iteration=0,
        link_key=None,
        link_tree=None,
        db_key=None,
    ):
        if self.artifact_db:
            item = LinkArtifact(
                key,
                artifact_path,
                link_iteration=link_iteration,
                link_key=link_key,
                link_tree=link_tree,
            )
            item.tree = tree
            item.iter = iter
            item.db_key = db_key or (name + "_" + key)
            self.artifact_db.store_artifact(
                item.db_key,
                item.to_dict(),
                tree=item.tree,
                iter=iter,
                tag=tag,
                project=project,
            )

    def delete_artifact(
        self,
        item: Artifact,
        deletion_strategy: mlrun.common.schemas.artifact.ArtifactsDeletionStrategies = (
            mlrun.common.schemas.artifact.ArtifactsDeletionStrategies.metadata_only
        ),
        secrets: typing.Optional[dict] = None,
    ):
        self.artifact_db.del_artifact(
            key=item.db_key,
            project=item.project,
            tree=item.tree,
            uid=item.uid,
            iter=item.iter,
            deletion_strategy=deletion_strategy,
            secrets=secrets,
        )


def extend_artifact_path(artifact_path: str, default_artifact_path: str):
    artifact_path = str(artifact_path or "")
    if artifact_path and artifact_path.startswith("+/"):
        if not default_artifact_path:
            return artifact_path[len("+/") :]
        if not default_artifact_path.endswith("/"):
            default_artifact_path += "/"
        return default_artifact_path + artifact_path[len("+/") :]
    return artifact_path or default_artifact_path


def filename(key, format):
    if not format:
        return key
    return f"{key}.{format}"



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import typing

import mlrun.common.schemas.secret
import mlrun.datastore.datastore_profile
import mlrun.errors
import mlrun.model_monitoring.helpers
from mlrun.datastore.datastore_profile import DatastoreProfile

from .base import TSDBConnector


class ObjectTSDBFactory(enum.Enum):
    """Enum class to handle the different TSDB connector type values for storing real time metrics"""

    v3io_tsdb = "v3io-tsdb"
    tdengine = "tdengine"

    def to_tsdb_connector(
        self, project: str, profile: DatastoreProfile, **kwargs
    ) -> TSDBConnector:
        """
        Return a TSDBConnector object based on the provided enum value.
        :param project: The name of the project.
        :param profile: Datastore profile containing DSN and credentials for TSDB connection
        :return: `TSDBConnector` object.
        """

        if self == self.v3io_tsdb:
            if mlrun.mlconf.is_ce_mode():
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"{self.v3io_tsdb} is not supported in CE mode."
                )

            from .v3io.v3io_connector import V3IOTSDBConnector

            return V3IOTSDBConnector(project=project, **kwargs)

        # Assuming TDEngine connector if connector type is not V3IO TSDB.
        # Update these lines once there are more than two connector types.

        from .tdengine.tdengine_connector import TDEngineConnector

        return TDEngineConnector(project=project, profile=profile, **kwargs)

    @classmethod
    def _missing_(cls, value: typing.Any):
        """A lookup function to handle an invalid value.
        :param value: Provided enum (invalid) value.
        """
        valid_values = list(cls.__members__.keys())
        raise mlrun.errors.MLRunInvalidMMStoreTypeError(
            f"{value} is not a valid tsdb, please choose a valid value: %{valid_values}."
        )


def get_tsdb_connector(
    project: str,
    secret_provider: typing.Optional[typing.Callable[[str], str]] = None,
    profile: typing.Optional[mlrun.datastore.datastore_profile.DatastoreProfile] = None,
) -> TSDBConnector:
    """
    Get TSDB connector object.
    :param project:                 The name of the project.
    :param secret_provider:         An optional secret provider to get the connection string secret.
    :param profile:                 An optional profile to initialize the TSDB connector from.

    :return: ``TSDBConnector`` object. The main goal of this object is to handle different operations on the
             TSDB connector such as updating drift metrics or write application record result.
    :raise: ``MLRunNotFoundError`` if the user didn't set the TSDB datastore profile and didn't provide it through
            the ``profile`` parameter.
    :raise: ``MLRunInvalidMMStoreTypeError`` if the TSDB datastore profile is of an invalid type.
    """
    profile = profile or mlrun.model_monitoring.helpers._get_tsdb_profile(
        project=project, secret_provider=secret_provider
    )
    kwargs = {}
    if isinstance(profile, mlrun.datastore.datastore_profile.DatastoreProfileV3io):
        tsdb_connector_type = mlrun.common.schemas.model_monitoring.TSDBTarget.V3IO_TSDB
    elif isinstance(
        profile, mlrun.datastore.datastore_profile.DatastoreProfileTDEngine
    ):
        tsdb_connector_type = mlrun.common.schemas.model_monitoring.TSDBTarget.TDEngine
    else:
        extra_message = (
            ""
            if profile
            else " by using `project.set_model_monitoring_credentials` API"
        )
        raise mlrun.errors.MLRunInvalidMMStoreTypeError(
            "You must provide a valid TSDB datastore profile"
            f"{extra_message}. "
            f"Found an unexpected profile of class: {type(profile)}"
        )

    # Get connector type value from ObjectTSDBFactory enum class
    tsdb_connector_factory = ObjectTSDBFactory(tsdb_connector_type)

    # Convert into TSDB connector object
    return tsdb_connector_factory.to_tsdb_connector(
        project=project, profile=profile, **kwargs
    )
